from rest_framework.reverse import reverse

from wbcore.metadata.configs.endpoints import EndpointViewConfig


class PlayerTeamEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> str:
        endpoint: str = super().get_create_endpoint(**kwargs)
        if team_id := self.view.kwargs.get("team_id"):
            endpoint += f"?team={team_id}"
        return endpoint

    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:player-team-list", args=[self.view.kwargs["team_id"]], request=self.request)


class MatchStadiumEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:match-stadium-list", args=[self.view.kwargs["stadium_id"]], request=self.request)


class MatchLeagueEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:match-league-list", args=[self.view.kwargs["league_id"]], request=self.request)


class TeamStadiumEndpointConfig(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs) -> str:
        endpoint: str = super().get_create_endpoint(**kwargs)
        if stadium_id := self.view.kwargs.get("stadium_id"):
            endpoint += f"?stadium={stadium_id}"
        return endpoint

    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:team-stadium-list", args=[self.view.kwargs["stadium_id"]], request=self.request)


class LeagueSportEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:league-sport-list", args=[self.view.kwargs["sport_id"]], request=self.request)


class EventTypeSportEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:eventtype-sport-list", args=[self.view.kwargs["sport_id"]], request=self.request)


class EventEndpointConfig(EndpointViewConfig):
    def get_delete_endpoint(self, **kwargs) -> None:
        return None


class LeagueStatisticsEndpointConfig(EndpointViewConfig):
    def get_delete_endpoint(self, **kwargs) -> None:
        return None

    def get_create_endpoint(self, **kwargs) -> None:
        return None

    def get_instance_endpoint(self, **kwargs) -> None:
        return None


class EventMatchEndpointConfig(EventEndpointConfig):
    def get_endpoint(self, **kwargs) -> str:
        return reverse("example_app:event-match-list", args=[self.view.kwargs["match_id"]], request=self.request)

    def get_create_endpoint(self, **kwargs) -> None:
        return None


class TeamResultsEndpointConfig(EndpointViewConfig):
    def get_delete_endpoint(self, **kwargs) -> None:
        return None

    def get_instance_endpoint(self, **kwargs) -> None:
        return None

    def get_create_endpoint(self, **kwargs) -> None:
        return None



import pytest
from django.contrib.contenttypes.models import ContentType
from wbcore.contrib.guardian.models.mixins import (
    assign_object_permissions_for_user_as_task,
    assign_user_permissions_for_object_as_task,
)


class TestAssignUserPermissionsForObjectAsTask:
    def test_called(self, mocker):
        content_type_class = mocker.patch("wbcore.contrib.guardian.models.mixins.ContentType")
        content_type = mocker.Mock()
        model_class = mocker.Mock()
        instance = mocker.Mock()
        model_class.objects.get.return_value = instance
        content_type.model_class.return_value = model_class
        content_type_class.objects.get.return_value = content_type

        assign_user_permissions_for_object_as_task(1, 1)

        instance.reload_permissions.assert_called_once_with(prune_existing=True)

    def test_called_with_prune_existing(self, mocker):
        content_type_class = mocker.patch("wbcore.contrib.guardian.models.mixins.ContentType")
        content_type = mocker.Mock()
        model_class = mocker.Mock()
        instance = mocker.Mock()
        model_class.objects.get.return_value = instance
        content_type.model_class.return_value = model_class
        content_type_class.objects.get.return_value = content_type

        assign_user_permissions_for_object_as_task(1, 1, prune_existing=False)

        instance.reload_permissions.assert_called_once_with(prune_existing=False)

    def test_not_called_without_model_class(self, mocker):
        content_type_class = mocker.patch("wbcore.contrib.guardian.models.mixins.ContentType")
        content_type = mocker.Mock()
        model_class = mocker.Mock()
        instance = mocker.Mock()
        model_class.objects.get.return_value = instance
        content_type.model_class.return_value = None
        content_type_class.objects.get.return_value = content_type

        assign_user_permissions_for_object_as_task(1, 1)

        instance.reload_permissions.assert_not_called()

    @pytest.mark.django_db
    def test_not_called_with_instance(self, user):
        with pytest.raises(
            AttributeError
        ):  # this means it crashed because the method reload_permissions does not exist which is fine
            assign_user_permissions_for_object_as_task(ContentType.objects.get_for_model(user).id, user.id)

    @pytest.mark.django_db
    def test_not_called_without_instance(self, user, mocker):
        user.reload_permissions = mocker.Mock()
        spy = mocker.spy(user, "reload_permissions")
        assign_user_permissions_for_object_as_task(ContentType.objects.get_for_model(user).id, user.id + 1)
        assert spy.call_count == 0


class TestAssignObjectPermissionsForUserAsTask:
    def test_called(self, mocker):
        user_class = mocker.patch("wbcore.contrib.guardian.models.mixins.User")
        user = mocker.Mock()

        user_class.objects.get.return_value = user

        get_inheriting_subclasses = mocker.patch("wbcore.contrib.guardian.models.mixins.get_inheriting_subclasses")
        get_inheriting_subclasses.return_value = [user_class]
        reload_permissions = mocker.patch("wbcore.contrib.guardian.models.mixins.reload_permissions")

        assign_object_permissions_for_user_as_task(user_id=1)

        reload_permissions.assert_called_once()



from rest_framework.reverse import reverse

from wbcore.contrib.directory.models import ClientManagerRelationship
from wbcore.metadata.configs.endpoints import EndpointViewConfig
from wbcore.permissions.shortcuts import is_internal_user


class RelationshipEntryModelEndpoint(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:entry-relationship-list",
            args=[
                self.view.kwargs["entry_id"],
            ],
            request=self.request,
        )


class ClientManagerEndpoint(EndpointViewConfig):
    def get_create_endpoint(self, **kwargs):
        base_url = "wbcore:directory:clientmanagerrelationship-list"
        filter_url = ""
        if "client" in self.request.GET:
            entry_id = self.request.GET["client"]
            filter_url = f"?client={entry_id}"
        if "relationship_manager" in self.request.GET:
            person_id = self.request.GET["relationship_manager"]
            filter_url = f"?relationship_manager={person_id}"
        return f"{reverse(base_url, args=[], request=self.request)}{filter_url}"

    def get_delete_endpoint(self, **kwargs):
        if "pk" in self.view.kwargs:
            cmr_obj = ClientManagerRelationship.objects.get(id=self.view.kwargs["pk"])
            if cmr_obj.status not in [
                ClientManagerRelationship.Status.DRAFT,
                ClientManagerRelationship.Status.REMOVED,
            ]:
                return None
        return super().get_delete_endpoint(**kwargs)


class EmployeeEmployerEndpointConfig(EndpointViewConfig):
    PK_FIELD = "employee"

    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:employer-employee-list", args=[self.view.kwargs["employer_id"]], request=self.request
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbcore:directory:person-list", args=[], request=self.request)

    def get_update_endpoint(self, **kwargs):
        return super().get_endpoint()


class EmployerEmployeeEndpointConfig(EndpointViewConfig):
    PK_FIELD = "employer"

    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcore:directory:employee-employer-list", args=[self.view.kwargs["employee_id"]], request=self.request
        )

    def get_instance_endpoint(self, **kwargs):
        return reverse("wbcore:directory:company-list", args=[], request=self.request)

    def get_update_endpoint(self, **kwargs):
        return super().get_endpoint()


class UserIsClientEndpointConfig(EndpointViewConfig):
    PK_FIELD = "relationship_manager"

    def get_endpoint(self, **kwargs):
        return None

    def get_instance_endpoint(self, **kwargs):
        if is_internal_user(self.request.user, include_superuser=True):
            return reverse("wbcore:directory:person-list", request=self.request)
        return None



import django_filters

from wbcore.filters.mixins import WBCoreFilterMixin
from wbcore.serializers import WBCoreType


class NumberFilter(WBCoreFilterMixin, django_filters.NumberFilter):
    filter_type = "number"

    def __init__(self, precision: int = 0, percent: bool = False, disable_formatting: bool = False, *args, **kwargs):
        self.precision = precision
        self.percent = percent
        self.disable_formatting = disable_formatting
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["precision"] = self.precision
        if self.percent:  # TODO: Discuss with Christoph if this is necessary like this
            lookup_expr["input_properties"]["type"] = WBCoreType.PERCENT.value
            lookup_expr["input_properties"]["precision"] = max(self.precision - 2, 0)
        lookup_expr["input_properties"]["disable_formatting"] = self.disable_formatting
        return representation, lookup_expr

    def filter(self, qs, value):
        if self.percent and value is not None:
            value /= 100
        return super().filter(qs, value)


class YearFilter(NumberFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.disable_formatting = True


class RangeSelectFilter(NumberFilter):
    filter_type = "rangeselect"

    def __init__(self, precision=2, *args, **kwargs):
        self.precision = precision
        self.color = kwargs.pop("color", "rgb(133, 144, 162)")
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        representation, lookup_expr = super().get_representation(request, name, view)
        lookup_expr["input_properties"]["color"] = self.color
        return representation, lookup_expr



from unittest.mock import MagicMock, patch

import pytest
from wbcore.contrib.i18n.viewsets import ModelTranslateMixin


class TestModel:
    """A simple test model for testing the ModelTranslateMixin."""

    id = 1


@pytest.fixture
def viewset():
    """Create a viewset instance for testing."""

    class TestViewSet(ModelTranslateMixin):
        def get_object(self):
            return TestModel()

    return TestViewSet()


@patch("wbcore.contrib.i18n.viewsets.translate_model_as_task")
@patch("wbcore.contrib.i18n.viewsets.ContentType.objects.get_for_model")
def test_auto_translate_success(mock_get_for_model, mock_translate_task, viewset, rf):
    """Test that auto_translate initiates a translation task and returns a success response."""
    # Setup
    mock_content_type = MagicMock()
    mock_content_type.id = 123
    mock_get_for_model.return_value = mock_content_type

    mock_translate_task.delay = MagicMock()

    request = rf.post("/auto-translate/")
    request.data = {}

    # Execute
    response = viewset.auto_translate(request)

    # Assert
    mock_get_for_model.assert_called_once()
    mock_translate_task.delay.assert_called_once_with(123, 1, False)

    assert response.status_code == 200
    assert response.data == {"__notification": "The translation started in the background."}


@patch("wbcore.contrib.i18n.viewsets.translate_model_as_task")
def test_auto_translate_no_object(mock_translate_task, viewset, rf):
    """Test that auto_translate returns an error response when no object is found."""
    # Setup
    viewset.get_object = MagicMock(return_value=None)
    mock_translate_task.delay = MagicMock()

    request = rf.post("/auto-translate/")
    request.data = {}
    # Execute
    response = viewset.auto_translate(request)

    # Assert
    mock_translate_task.delay.assert_not_called()

    # The viewset code uses status.HTTP_400_BAD_REQUEST
    assert response.status_code == 400
    assert "non_field_errors" in response.data
    assert response.data["non_field_errors"] == ["The URL was malformatted."]



import pytest
from pytest_mock import MockerFixture
from wbcore.contrib.workflow.models import DecisionStep, Transition


class TestDecisionStep:
    @pytest.fixture
    def mocked_instances(self, mocker: MockerFixture):
        return mocker.MagicMock(first_name="John")

    @pytest.fixture
    def mocked_decision_step(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=DecisionStep)

    def test_get_first_valid_transition(
        self,
        mocker: MockerFixture,
        mocked_process_step,
        mocked_instances,
        mocked_decision_step,
    ):
        # Arrange
        mocked_process_step.get_instance.return_value = mocked_instances
        mocked_process_step.step = mocked_decision_step
        valid_transition1 = mocker.MagicMock(spec=Transition)
        valid_transition2 = mocker.MagicMock(spec=Transition)
        invalid_transition = mocker.MagicMock(spec=Transition)
        mocked_decision_step.get_outgoing_transitions.return_value = [
            invalid_transition,
            valid_transition1,
            valid_transition2,
        ]
        invalid_transition.all_conditions_satisfied.return_value = False
        valid_transition1.all_conditions_satisfied.return_value = True
        valid_transition2.all_conditions_satisfied.return_value = True
        # Act
        result = DecisionStep.get_first_valid_transition(mocked_decision_step, mocked_process_step)
        # Assert
        assert result in [valid_transition1, valid_transition2]
        invalid_transition.all_conditions_satisfied.assert_called_once_with(mocked_process_step)
        valid_transition1.all_conditions_satisfied.assert_called_once_with(mocked_process_step)
        valid_transition2.all_conditions_satisfied.assert_not_called()

    def test_get_first_valid_transition_no_transitions(
        self, mocker, mocked_process_step, mocked_instances, mocked_decision_step
    ):
        # Arrange
        mocked_process_step.get_instance.return_value = mocked_instances
        mocked_process_step.step = mocked_decision_step
        invalid_transition = mocker.MagicMock(spec=Transition)
        mocked_decision_step.get_outgoing_transitions.return_value = [invalid_transition]
        invalid_transition.all_conditions_satisfied.return_value = False
        # Act
        result = DecisionStep.get_first_valid_transition(mocked_decision_step, mocked_process_step)
        # Assert
        assert result is None
        invalid_transition.all_conditions_satisfied.assert_called_once_with(mocked_process_step)

    def test_run(self, mocker: MockerFixture, mocked_process_step, mocked_decision_step):
        # Arrange
        mocked_start_next_step = mocker.patch("wbcore.contrib.workflow.models.step.Step.start_next_step")
        mocked_transition = mocker.MagicMock(spec=Transition)
        mocked_decision_step.get_first_valid_transition.return_value = mocked_transition
        # Act
        DecisionStep.run(mocked_decision_step, mocked_process_step)
        # Assert
        mocked_decision_step.get_first_valid_transition.assert_called_once_with(mocked_process_step)
        mocked_start_next_step.assert_called_once_with(mocked_process_step, mocked_transition)

    def test_run_failed(self, mocker: MockerFixture, mocked_process_step, mocked_decision_step):
        # Arrange
        mocked_activate_step = mocker.patch("wbcore.contrib.workflow.models.step.activate_step.delay")
        mocked_decision_step.get_first_valid_transition.return_value = None
        mocked_set_failed = mocker.patch.object(mocked_decision_step, "set_failed")
        # Act
        DecisionStep.run(mocked_decision_step, mocked_process_step)
        # Assert
        mocked_decision_step.get_first_valid_transition.assert_called_once_with(mocked_process_step)
        mocked_set_failed.assert_called_once_with(
            mocked_process_step, "No valid outgoing transition found for this step!"
        )
        mocked_activate_step.assert_not_called()



from django.apps import apps
from django.db.models.signals import pre_migrate
from pytest_factoryboy import register
from wbcore.contrib.color.factories import ColorGradientFactory
from wbcore.contrib.geography.tests.signals import app_pre_migration

register(ColorGradientFactory)


pre_migrate.connect(app_pre_migration, sender=apps.get_app_config("geography"))



from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.color import Color
from selenium.webdriver.support.wait import WebDriverWait

from .e2e_helper_methods import find_element


def does_background_color_match(element: WebElement, expected_color: str) -> bool:
    """Check if the background color of an element matches the expected color.

    Args:
        element (WebElement): The element to check the background of.
        expected_color (str): The expected color.

    Returns:
        bool: True if the background color is as expected
    """
    element_rgb = element.value_of_css_property("background-color")
    element_rgb = Color.from_string(element_rgb).rgb
    return element_rgb == expected_color


def is_text_visible(driver: WebDriver, text: str) -> bool:
    """Check if the searched text is visible on the current page.

    Args:
        driver (WebDriver): The Selenium webdriver.
        text (str): Searched text.

    Returns:
        bool: True if the text is visible
    """
    return True if find_element(driver, f"//*[text()='{text}']") else False


def is_string_not_visible(driver: WebDriver, string: str) -> bool:
    """Check if the searched text is not visible on the current page.
    Sometimes selenium is faster, that the page can refresh. So this method can wait up to 10 seconds for a string to not be visible.

    Args:
        driver (WebDriver): The Selenium webdriver.
        text (str): Searched text.

    Returns:
        bool: True if the text is not visible
    """
    try:
        WebDriverWait(driver, 5).until(
            expected_conditions.invisibility_of_element_located((By.XPATH, f"//*[text()='{string}']"))
        )
        return True
    except TimeoutException:
        return False


def is_tag_visible(driver: WebDriver, tag_label: str) -> bool:
    """Check if a tag with a certain label is visible on the current page.

    Args:
        driver (WebDriver): The Selenium webdriver.
        text (str): The label text.


    Returns:
        bool: True if the label is visible
    """
    return True if find_element(driver, f"//*[@class='tag-label' and text()='{tag_label}']") else False


def is_tag_not_visible(driver: WebDriver, tag_label: str) -> bool:
    """Check if a tag with a certain label is not visible on the current page.
    Sometimes selenium is faster, that the page can refresh. So this method can wait up to 10 seconds for a tag to not be visible.

    Args:
        driver (WebDriver): The Selenium webdriver.
        text (str): The label text.

    Returns:
        bool: True if the tag is not visible
    """
    try:
        WebDriverWait(driver, 2.5).until(
            expected_conditions.invisibility_of_element_located(
                (By.XPATH, f"//*[@class='tag-label' and text()='{tag_label}']")
            )
        )
        return True
    except TimeoutException:
        return False


def is_error_visible(driver: WebDriver):
    """Check if an error element is visible on the current page.

    Args:
        driver (WebDriver): The Selenium webdriver.

    Returns:
        bool: True if an error element is visible
    """
    error_element = find_element(driver, "//div[contains(@type, 'error')]", 2.5)
    saving_failed_hint = find_element(driver, "//div[contains(@class, 'task-dropper-content')]", 2.5)

    if saving_failed_hint and saving_failed_hint.is_displayed():
        WebDriverWait(driver, 10).until_not(
            expected_conditions.invisibility_of_element_located(
                (By.XPATH, "//div[contains(@class, 'task-dropper-content')]")
            )
        )
    return error_element is not None


def is_counter_as_expected(driver: WebDriver, count: int):
    """Check if list counter displays the right number.

    Args:
        driver (WebDriver): The Selenium webdriver.

    Returns:
        bool: True if the counter is right.
    """
    find_element(driver, f"//span[@class='ag-status-name-value-value' and text()='{count}']")
    return find_element is not None



def register(section=None, weight=0):
    """
    Register the given serializer and the associated section to the shares site
    """
    from wbcore.shares.sites import share_site

    def _serializer_wrapper(serializer_class):
        share_site.serializers.append(serializer_class)
        if section:
            share_site.sections.append({"section": section, "weight": weight})
        return serializer_class

    return _serializer_wrapper



from ..configurations import configuration_registry

CompanyModelViewSet = configuration_registry.company_model_viewset
CompanyRepresentationViewSet = configuration_registry.company_representation_viewset
PersonModelViewSet = configuration_registry.person_model_viewset
PersonRepresentationViewSet = configuration_registry.person_representation_viewset

from .contacts import (
    AddressContactEntryViewSet,
    AddressContactViewSet,
    BankingContactEntryViewSet,
    BankingContactRepresentationViewSet,
    BankingContactViewSet,
    EmailContactEntryViewSet,
    EmailContactRepresentationViewSet,
    EmailContactViewSet,
    SocialMediaContactEntryViewSet,
    SocialMediaContactRepresentationViewSet,
    SocialMediaContactViewSet,
    TelephoneContactEntryViewSet,
    TelephoneContactRepresentationViewSet,
    TelephoneContactViewSet,
    WebsiteContactEntryViewSet,
    WebsiteContactRepresentationViewSet,
    WebsiteContactViewSet,
)
from .entries import (
    BankModelViewSet,
    EntryModelViewSet,
    EntryRepresentationViewSet,
    PersonInChargeRepresentationViewSet,
    UserIsManagerViewSet,
)
from .relationships import (
    ClientManagerRelationshipRepresentationViewSet,
    ClientManagerViewSet,
    EmployeeEmployerModelViewSet,
    EmployerEmployeeModelViewSet,
    EmployerEmployeeRelationshipModelViewSet,
    RelationshipModelEntryViewSet,
    RelationshipModelViewSet,
    RelationshipRepresentationViewSet,
    RelationshipTypeModelViewSet,
    RelationshipTypeRepresentationViewSet,
    SystemEmployeeModelViewSet,
    UserIsClientViewSet,
)
from .utils import (
    CompanyTypeModelViewSet,
    CompanyTypeRepresentationViewSet,
    CustomerStatusModelViewSet,
    CustomerStatusRepresentationViewSet,
    PositionModelViewSet,
    PositionRepresentationViewSet,
    SpecializationModelViewSet,
    SpecializationRepresentationViewSet,
)



from datetime import timedelta

from celery import shared_task
from django.conf import settings
from django.contrib.auth import user_logged_in
from django.db import models
from django.db.models import DateTimeField, OuterRef, Subquery
from django.dispatch import receiver
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from .users import User

ACCESS_TOKEN_LIFETIME = settings.SIMPLE_JWT["ACCESS_TOKEN_LIFETIME"].seconds


class UserActivity(models.Model):
    SUCCESS = _("Success")
    FAILED = _("Failed")

    ACTIVITY_STATUS = ((SUCCESS, _("Success")), (FAILED, _("Failed")))

    class Type(models.TextChoices):
        LOGIN = "LOGIN", _("Login")

    IP = models.GenericIPAddressField(null=True, blank=True)
    date = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(
        "authentication.User", related_name="login_activities", null=True, on_delete=models.SET_NULL
    )
    status = models.CharField(max_length=8, default=SUCCESS, choices=ACTIVITY_STATUS, null=True, blank=True)
    jti = models.CharField(max_length=32, default="", blank=True)
    type = models.CharField(max_length=16, default=Type.LOGIN, choices=Type.choices, null=True, blank=True)
    user_agent_info = models.CharField(max_length=255)
    latest_refresh = models.DateTimeField(null=True, blank=True)

    def __str__(self) -> str:
        return _("Activity of {user} at {date}").format(user=str(self.user), date=self.date)

    class Meta:
        verbose_name = _("User Login Activity")
        verbose_name_plural = _("User Login Activities")

    @classmethod
    def get_latest_login_datetime_subquery(cls, profile_name="pk", use_user=False):
        """
        Create a subquery to retrieve the latest login datetime information

        Arguments:
            profile_name {str} -- The OuterRef field for user__profile
        Returns:
            Subquery -- Subquery for latest login detail
        """

        if use_user:
            qs = UserActivity.objects.filter(user__id=OuterRef("id"))
        else:
            qs = UserActivity.objects.filter(user__profile__id=OuterRef(profile_name))
        return Subquery(qs.order_by("-latest_refresh").values("latest_refresh")[:1], output_field=DateTimeField())

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:authentication:useractivity"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{user}}: {{date}}-{{latest_refresh}}"


@receiver(user_logged_in)
def log_user_logged_in_success(sender, user, request, **kwargs):
    """
    Signal triggerd whenever a user log into a new session. Create a new UserActivity for that user
    """
    user_agent_info = (request.META.get("HTTP_USER_AGENT", "<unknown>")[:255],)
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[0]
    else:
        ip = request.META.get("REMOTE_ADDR")

    access_token_lifespan = settings.SIMPLE_JWT["ACCESS_TOKEN_LIFETIME"].seconds
    state = UserActivity.SUCCESS if kwargs.get("authentication_status", True) else UserActivity.FAILED
    if jti := kwargs.get("jti", None):
        # TODO: What if request is None?
        UserActivity.objects.create(
            IP=ip,
            user=user,
            user_agent_info=user_agent_info,
            status=state,
            jti=jti,
            latest_refresh=timezone.now() + timedelta(seconds=access_token_lifespan),
        )


@shared_task()
def refresh_user_activity(user_id, refresh_time, jti):
    """
    Signal triggerd whenever a refresh token is submit from the frontend. Update latest_refresh to the associated UserActivity.
    """
    user = User.objects.get(id=user_id)
    user_activity, created = UserActivity.objects.get_or_create(user=user, jti=jti)
    if created:
        elapse = (refresh_time - user_activity.latest_refresh).seconds
        access_token_lifespan = settings.SIMPLE_JWT["ACCESS_TOKEN_LIFETIME"].seconds
        if elapse > access_token_lifespan:
            user_activity.latest_refresh = refresh_time
            user_activity.save()



from datetime import timedelta
from decimal import Decimal

import pytest
from django.utils import timezone
from pytest_mock import MockerFixture

from wbcore.contrib.currency.models import Currency, CurrencyFXRates


@pytest.mark.model_tests
@pytest.mark.with_db
@pytest.mark.django_db
class TestSpecificModels:
    @pytest.fixture
    def base_currency(self, mocker: MockerFixture):
        base_currency = Currency()
        mocker.patch.object(base_currency, "key", "ABC")
        return base_currency

    @pytest.fixture
    def other_currency(self, mocker: MockerFixture):
        base_currency = Currency()
        mocker.patch.object(base_currency, "key", "XYZ")
        return base_currency

    @pytest.fixture
    def base_fx_rate(self, mocker: MockerFixture):
        base_fx_rate = CurrencyFXRates()
        mocker.patch.object(base_fx_rate, "value", 0.85)
        return base_fx_rate

    @pytest.fixture
    def other_fx_rate(self, mocker: MockerFixture):
        other_fx_rate = CurrencyFXRates()
        mocker.patch.object(other_fx_rate, "value", 1.15)
        return other_fx_rate

    @pytest.mark.parametrize("exact_lookup", [True, False])
    def test_convert(
        self,
        exact_lookup,
        mocker: MockerFixture,
        base_currency,
        other_currency,
        base_fx_rate,
        other_fx_rate,
    ):
        # Arrange
        today = timezone.now().date()
        if exact_lookup:
            mocker.patch(
                "wbcore.contrib.currency.models.CurrencyFXRates.objects.get",
                side_effect=[base_fx_rate, other_fx_rate],
            )
        else:
            mock_filter = mocker.patch("wbcore.contrib.currency.models.CurrencyFXRates.objects.filter")
            mock_filter.return_value.order_by.return_value.first.side_effect = [
                base_fx_rate,
                other_fx_rate,
            ]

        # Act
        new_fx_rates = base_currency.convert(today, other_currency, exact_lookup)

        # Assert
        assert new_fx_rates == (1 / base_fx_rate.value) * other_fx_rate.value

    def test_convert_with_date_today_exists(self, currency_factory, currency_fx_rates_factory):
        currency = currency_factory(key="EUR")
        other_currency = currency_factory(key="CHF")
        today = timezone.now().date()
        fx_rates = currency_fx_rates_factory(date=today, currency=currency)
        other_fx_rates = currency_fx_rates_factory(date=today, currency=other_currency)
        new_fx_rates = currency.convert(fx_rates.date, other_currency)
        assert new_fx_rates == (1 / fx_rates.value) * other_fx_rates.value

    def test_convert_with_date_lte(self, currency_factory, currency_fx_rates_factory):
        date_test = timezone.now().date() - timedelta(days=1)
        currency = currency_factory(key="EUR")
        other_currency = currency_factory(key="CHF")
        today = timezone.now().date()
        fx_rates = currency_fx_rates_factory(date=date_test, currency=currency)
        other_fx_rates = currency_fx_rates_factory(date=date_test, currency=other_currency)
        new_fx_rates = currency.convert(today, other_currency)
        assert new_fx_rates == (1 / fx_rates.value) * other_fx_rates.value

    def test_get_fx_rates_subquery_normal_date(self, currency_fx_rates_factory):
        today = timezone.now().date()
        fx_rates = currency_fx_rates_factory(date=today)

        qs = CurrencyFXRates.objects.annotate(fx_rate=fx_rates.get_fx_rates_subquery(today, currency="currency"))

        fx_rate = [fx_r.get("fx_rate") for fx_r in qs.filter(id=fx_rates.id).values("fx_rate")]
        assert len(fx_rate) == 1
        assert isinstance(fx_rate[0], Decimal)

    def test_get_fx_rates_subquery_outerref_date(self, currency_fx_rates_factory):
        today = timezone.now().date()
        fx_rates = currency_fx_rates_factory(date=today)
        qs = CurrencyFXRates.objects.annotate(fx_rate=fx_rates.get_fx_rates_subquery("date", currency="currency"))
        fx_rate = [fx_r.get("fx_rate") for fx_r in qs.filter(id=fx_rates.id).values("fx_rate")]
        assert len(fx_rate) == 1
        assert isinstance(fx_rate[0], Decimal)

    def test_get_fx_rates_subquery_for_two_currencies(self, currency_fx_rates_factory):
        today = timezone.now().date()
        fx_rates = currency_fx_rates_factory(date=today)
        qs = CurrencyFXRates.objects.annotate(
            fx_rate=fx_rates.get_fx_rates_subquery_for_two_currencies("date", "currency", "currency")
        )
        fx_rate = [fx_r.get("fx_rate") for fx_r in qs.filter(id=fx_rates.id).values("fx_rate")]
        assert len(fx_rate) == 1
        assert isinstance(fx_rate[0], Decimal)
        assert fx_rate[0] == Decimal("1.0")



from collections import defaultdict
from datetime import date, datetime, time, timedelta, timezone
from zoneinfo import ZoneInfo, available_timezones

from dateutil import rrule
from django.utils.dateparse import parse_date
from django.utils.timezone import localdate
from pandas.tseries.offsets import BDay

from wbcore.utils.date_builder import (
    BusinessDay,
    Hour,
    Minute,
    Month,
    MonthEnd,
    MonthStart,
    Now,
    Quarter,
    QuarterEnd,
    QuarterStart,
    Week,
    Year,
    YearStart,
)
from wbcore.utils.date_builder.components import Component


def get_quarter_from_date(d):
    return ((d.month - 1) // 3) + 1


def get_start_date_from_date(d):
    quarter = get_quarter_from_date(d)
    return date(d.year, quarter * 3 - 2, 1)


def get_end_date_from_date(d):
    quarter = get_quarter_from_date(d)
    return date(d.year + ((quarter * 3 + 1) // 12), (quarter * 3 + 1) % 12, 1) - timedelta(days=1)


def get_start_and_end_date_from_date(d):
    return get_start_date_from_date(d), get_end_date_from_date(d)


def current_quarter_date_start(field=None, request=None, view=None):
    return get_start_date_from_date(localdate())


def current_quarter_date_end(field=None, request=None, view=None):
    return get_end_date_from_date(localdate())


def current_quarter_date_interval(field, request, view):
    return (
        current_quarter_date_start(field, request, view),
        current_quarter_date_end(field, request, view),
    )


def financial_year_to_date(field, request, view):
    d = localdate()
    return (date(d.year - 1, 12, 31), d)


def current_year_date_start(field, request, view):
    d = localdate()
    return date(d.year, 1, 1)


def current_year_date_end(field, request, view):
    d = localdate()
    return date(d.year + 1, 1, 1) - timedelta(days=1)


def current_year_date_interval(field, request, view):
    return (
        current_year_date_start(field, request, view),
        current_year_date_end(field, request, view),
    )


def current_month_date_start(field, request, view):
    d = localdate()
    return date(d.year, d.month, 1)


def current_month_date_end(field=None, request=None, view=None):
    d = localdate()
    if d.month == 12:
        return date(d.year, 12, 31)
    return date(d.year, d.month + 1, 1) - timedelta(days=1)


def current_month_date_interval(field, request, view):
    return (
        current_month_date_start(field, request, view),
        current_month_date_end(field, request, view),
    )


def get_date_interval_from_request(
    request,
    request_type="GET",
    exclude_weekend=False,
    left_interval_inclusive=False,
    right_interval_inclusive=True,
    date_range_fieldname="date",
):
    """
    Parses a request and returns the start and end date from it.

    Parameters
    ----------
    request: The GET Request Object
    exclude_weekend: If True, exclude weekend
    left_interval_inclusive: If False, corresponds to left bound excluded "(" otherwise, inclusive "["
    right_interval_inclusive: If False, corresponds to right bound excluded ")" otherwise, inclusive "]"

    Returns
    -------
    Return a tuple in the form of (start_date, end_date). If either the start date or the end date is not present in the request None is returned in the tuple
    """

    start_identifier = ["start", "start_date", "from", "date_gte", "date__gte"]
    end_identifier = ["end", "end_date", "to", "date_lte", "date__lte"]
    params = request.GET if request_type == "GET" else request.POST
    start = None
    end = None
    if date_range_fieldname in params:
        if len(params.get(date_range_fieldname).split(",")) == 2:
            start, end = map(lambda x: None if x == "" else x, params.get(date_range_fieldname).split(","))
    if not start or not end:
        start = next(
            (params.get(identifier) for identifier in start_identifier if identifier in params),
            None,
        )
        end = next((params.get(identifier) for identifier in end_identifier if identifier in params), None)
    if start:
        start = parse_date(start)
        if start and exclude_weekend:
            if left_interval_inclusive:
                start = (start - timedelta(days=1) + BDay(1)).date()
            else:
                start = (start + timedelta(days=1) - BDay(1)).date()
    if end:
        end = parse_date(end)
        if end and exclude_weekend:
            if right_interval_inclusive:
                end = (end + timedelta(days=1) - BDay(1)).date()
            else:
                end = (end - timedelta(days=1) + BDay(1)).date()
    return start, end


def get_number_of_hours_between_dates(
    d1, d2, skip_weekends=True, list_public_holidays=False, hours_range=None, granularity=12
):
    if hours_range is None:
        hours_range = range(0, 23)

    def convert_days_from_hours(hours, granularity, hours_per_day):
        return int(hours / granularity) * granularity / hours_per_day

    rules = rrule.rruleset()

    byweekday_list = [rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR]
    if not skip_weekends:
        byweekday_list.extend([rrule.SA, rrule.SU])

    rules.rrule(
        rrule.rrule(
            freq=rrule.HOURLY,
            byweekday=byweekday_list,
            byhour=hours_range,
            dtstart=d1,
            until=d2,
        )
    )
    if list_public_holidays:
        for holiday in list_public_holidays:
            s1 = datetime(holiday.year, holiday.month, holiday.day, 0, 0, 0)
            s2 = datetime(holiday.year, holiday.month, holiday.day, 23, 59, 59)
            rules.exrule(rrule.rrule(rrule.HOURLY, dtstart=s1, until=s2))
    dates = defaultdict(int)
    for r in list(rules):
        dates[r.date()] += 1
    return {k: convert_days_from_hours(v, granularity, len(hours_range)) for k, v in dates.items()}


def shortcut(label: str, value_start: str | Component, value_end: str | Component) -> dict[str, str]:
    return {"label": label, "value": f"{str(value_start)},{str(value_end)}"}


financial_performance_shortcuts = [
    shortcut("Last Day", Now - BusinessDay, Now),
    shortcut("Last Week", Now - Week - BusinessDay, Now),
    shortcut("Month to Date", MonthStart - BusinessDay, Now),
    shortcut("Last Month", Now - Month - BusinessDay, Now),
    shortcut("Quarter to Date", QuarterStart - BusinessDay, Now),
    shortcut("Last Quarter", Now - Quarter - BusinessDay, Now),
    shortcut("Year to Date", YearStart - BusinessDay, Now),
    shortcut("Last 1 Year", Now - Year - BusinessDay, Now),
    shortcut("Last 2 Years", Now - Year(2) - BusinessDay, Now),
    shortcut("Last 3 Years", Now - Year(3) - BusinessDay, Now),
    shortcut("Last 5 Years", Now - Year(5) - BusinessDay, Now),
]

current_financial_month = MonthStart - BusinessDay, MonthEnd
current_financial_quarter = QuarterStart - BusinessDay, QuarterEnd

calendar_item_shortcuts = [
    shortcut("Last Minute", Now - Minute, Now),
    shortcut("Last 5 Minutes", Now - Minute(5), Now),
    shortcut("Last 10 Minutes", Now - Minute(10), Now),
    shortcut("Last 15 Minutes", Now - Minute(15), Now),
    shortcut("Last 30 Minutes", Now - Minute(30), Now),
    shortcut("Last Hour", Now - Hour, Now),
]


def get_next_day_timedelta(now: datetime | None = None) -> int:
    if not now:
        now = datetime.now()
    return (datetime.combine(now.date() + timedelta(days=1), time(0, 0, 0)) - now).seconds


def get_timezone_choices() -> list[tuple[str, str]]:
    now_utc = datetime.now(timezone.utc)
    tz_tuples = []  # a list of (timezone_name, timezone_name (UTC offset))
    for tz_name in sorted(available_timezones()):
        tz = ZoneInfo(tz_name)
        now_in_tz = now_utc.astimezone(tz)
        offset_str = now_in_tz.strftime("UTC%z")  # gives UTC+HHMM
        offset_str = offset_str[:-2] + ":" + offset_str[-2:]
        tz_tuples.append((tz_name, f"{tz_name} ({offset_str})"))
    return tz_tuples



import os

from configurations import values


class Base:
    DEBUG = values.BooleanValue(True, environ_prefix=None)
    SECRET_KEY = values.Value("THIS-IS-NOT-A-SECRET-KEY", environ_prefix=None)
    SITE_ID = values.IntegerValue(1, environ_prefix=None)

    @property
    def PROJECT_NAME(self):  # noqa
        if settings_module := os.environ.get("DJANGO_SETTINGS_MODULE", None):
            return settings_module.split(".")[0]
        return None



from django.utils.translation import gettext as _
from rest_framework.reverse import reverse

from wbcore.contrib.directory.models import ClientManagerRelationship
from wbcore.contrib.directory.permissions import IsClientManagerRelationshipAdmin
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig


class EmployerEmployeeRelationshipButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        if not self.new_mode:
            if employer_id := self.view.kwargs.get("employer_id", None):
                base_url = reverse("wbcore:directory:person-list", args=[], request=self.request)
                return {
                    bt.WidgetButton(
                        endpoint=f"{base_url}?employers={employer_id}",
                        label=_("New Person"),
                        icon=WBIcon.PERSON_ADD.icon,
                        new_mode=True,
                    )
                }
            if employee_id := self.view.kwargs.get("employee_id", None):
                base_url = reverse("wbcore:directory:company-list", args=[], request=self.request)
                return {
                    bt.WidgetButton(
                        endpoint=f"{base_url}?employees={employee_id}",
                        label=_("New Company"),
                        icon=WBIcon.ADD.icon,
                        new_mode=True,
                    )
                }
        return {}


class ClientManagerRelationshipButtonConfig(ButtonViewConfig):
    def get_custom_buttons(self):
        buttons = set()
        if (
            IsClientManagerRelationshipAdmin().has_permission(self.request, self.view)
            and ClientManagerRelationship.objects.filter(
                status__in=[
                    ClientManagerRelationship.Status.PENDINGADD,
                    ClientManagerRelationship.Status.PENDINGREMOVE,
                ]
            ).exists()
        ):
            buttons.add(
                bt.ActionButton(
                    method=RequestType.PATCH,
                    identifiers=("directory:clientmanagerrelationship",),
                    endpoint=reverse("wbcore:directory:clientmanagerrelationship-approveallpendingrequests"),
                    label=_("Approve All Pending Requests"),
                    icon=WBIcon.APPROVE.icon,
                    description_fields=_(
                        "<p> Are you sure you want to approve all pending (add and remove) requests ? </p>"
                    ),
                    title=_("Approve All Pending Requests"),
                    action_label=_("Approved"),
                )
            )
        return buttons



import warnings

warnings.warn(
    "The 'wbcore.pandas' module is deprecated and will be removed in a future version. "
    "Please use 'wbcore.contrib.pandas' instead.",
    DeprecationWarning,
    stacklevel=2,
)
from wbcore.contrib.pandas.filterset import PandasFilterSetMixin  # noqa



from django.db.models import QuerySet
from django.utils.translation import gettext_lazy as _
from psycopg.types.range import DateRange, TimestamptzRange

from wbcore import filters
from wbcore.contrib.example_app.models import (
    Event,
    EventType,
    Match,
    Sport,
    SportPerson,
)
from wbcore.filters.defaults import current_year_date_range


class EventMatchFilter(filters.FilterSet):
    person = filters.ModelMultipleChoiceFilter(
        label=_("Persons"),
        queryset=SportPerson.objects.all(),
        endpoint=SportPerson.get_representation_endpoint(),
        value_key=SportPerson.get_representation_value_key(),
        label_key=SportPerson.get_representation_label_key(),
    )
    minute__gte = filters.NumberFilter(label=_("Minute"), lookup_expr="gte", field_name="minute")
    minute__lte = filters.NumberFilter(label=_("Minute"), lookup_expr="lte", field_name="minute")
    event_type = filters.ModelMultipleChoiceFilter(
        label=_("Event Types"),
        queryset=SportPerson.objects.all(),
        endpoint=SportPerson.get_representation_endpoint(),
        value_key=SportPerson.get_representation_value_key(),
        label_key=SportPerson.get_representation_label_key(),
    )

    class Meta:
        model = Event
        fields = {}


class EventFilter(EventMatchFilter):
    match = filters.ModelChoiceFilter(
        label=_("Match"),
        queryset=Match.objects.all(),
        endpoint=Match.get_representation_endpoint(),
        value_key=Match.get_representation_value_key(),
        label_key=Match.get_representation_label_key(),
    )


class EventTypeSportFilter(filters.FilterSet):
    points__gte = filters.NumberFilter(label=_("Points"), lookup_expr="gte", field_name="points")
    points__lte = filters.NumberFilter(label=_("Points"), lookup_expr="lte", field_name="points")

    class Meta:
        model = EventType
        fields = {
            "name": ["exact", "icontains"],
            "icon": ["exact", "icontains"],
        }


class EventTypeFilter(EventTypeSportFilter):
    sport = filters.ModelMultipleChoiceFilter(
        label=_("Sports"),
        queryset=Sport.objects.all(),
        endpoint=Sport.get_representation_endpoint(),
        value_key=Sport.get_representation_value_key(),
        label_key=Sport.get_representation_label_key(),
    )


class PlayerStatisticsChartFilter(filters.FilterSet):
    minute__gte = filters.NumberFilter(label=_("Minute"), lookup_expr="gte", field_name="minute")
    minute__lte = filters.NumberFilter(label=_("Minute"), lookup_expr="lte", field_name="minute")
    period = filters.DateRangeFilter(
        label=_("Period"), method="filter_by_period", initial=current_year_date_range, required=True
    )

    def filter_by_period(self, queryset: QuerySet[Event], name, value: DateRange) -> QuerySet[Event]:
        return queryset.filter(match__date_time__contained_by=TimestamptzRange(value.lower, value.upper))

    class Meta:
        model = Event
        fields = {}



import logging
from collections import OrderedDict
from contextlib import suppress
from copy import copy

from django.contrib.postgres.fields import DateRangeField, DateTimeRangeField
from django.core.exceptions import FieldError
from django.db import models
from django.db.models import GeneratedField
from django.db.models.fields.related import ManyToManyRel, ManyToOneRel, OneToOneRel
from django_filters.filterset import FilterSetMetaclass, remote_queryset, settings
from django_filters.rest_framework import FilterSet as DjangoFilterSet

from wbcore.filters import fields
from wbcore.filters.fields.multiple_lookups import MultipleLookupFilter
from wbcore.models.fields import YearField
from wbcore.signals.filters import add_filters

from .utils import check_required_filters

logger = logging.getLogger(__name__)


def _is_number(field):
    return (
        issubclass(field.__class__, models.DecimalField)
        or issubclass(field.__class__, models.FloatField)
        or issubclass(field.__class__, models.IntegerField)
    )


class CustomFilterSetMetaClass(FilterSetMetaclass):
    def __new__(cls, *args, **kwargs):
        new_class = super().__new__(cls, *args, **kwargs)
        if _meta := getattr(new_class, "Meta", None):
            for parent_field_name, child_fields in getattr(_meta, "flatten_fields", dict()).items():
                if remote_field := getattr(_meta.model._meta.get_field(parent_field_name), "remote_field", None):
                    for child_field_name, lookup_exps in child_fields.items():
                        child_field = remote_field.model._meta.get_field(child_field_name)
                        if _is_number(child_field):
                            nested_field_name = f"{parent_field_name}__{child_field_name}"
                            for lookup_exp in lookup_exps:
                                new_class.base_filters[f"{nested_field_name}__{lookup_exp}"] = fields.NumberFilter(
                                    field_name=nested_field_name,
                                    lookup_expr=lookup_exp,
                                    precision=2,
                                    label=child_field.verbose_name,
                                )

                        # TODO extend with other filter fields type
        return new_class

    @classmethod
    def get_declared_filters(cls, bases, attrs):
        filters = super().get_declared_filters(bases, attrs)

        multi_filters: list[tuple[str, MultipleLookupFilter]] = [
            (filter_name, attrs.pop(filter_name))
            for filter_name, obj in list(attrs.items())
            if isinstance(obj, MultipleLookupFilter)
        ]

        for field_name, filter in multi_filters:
            filters |= filter.get_filters(field_name)

        return filters


class FilterSet(DjangoFilterSet, metaclass=CustomFilterSetMetaClass):
    DEFAULT_EXCLUDE_FILTER_LOOKUP: str = "exclude"
    FILTER_DEFAULTS = {
        models.BooleanField: {"filter_class": fields.BooleanFilter},
        models.NullBooleanField: {"filter_class": fields.BooleanFilter},
        models.CharField: {"filter_class": fields.CharFilter},
        models.TextField: {"filter_class": fields.CharFilter},
        models.SlugField: {"filter_class": fields.CharFilter},
        models.EmailField: {"filter_class": fields.CharFilter},
        models.FilePathField: {"filter_class": fields.CharFilter},
        models.UUIDField: {"filter_class": fields.CharFilter},
        models.URLField: {"filter_class": fields.CharFilter},
        models.GenericIPAddressField: {"filter_class": fields.CharFilter},
        models.CommaSeparatedIntegerField: {"filter_class": fields.CharFilter},
        models.DateField: {"filter_class": fields.DateFilter},
        models.DateTimeField: {"filter_class": fields.DateTimeFilter},
        models.TimeField: {"filter_class": fields.TimeFilter},
        models.IntegerField: {"filter_class": fields.NumberFilter},
        models.FloatField: {"filter_class": fields.NumberFilter},
        models.DecimalField: {"filter_class": fields.NumberFilter},
        DateTimeRangeField: {"filter_class": fields.DateTimeRangeFilter},
        DateRangeField: {"filter_class": fields.DateRangeFilter},
        YearField: {"filter_class": fields.YearFilter},
        # models.DurationField: {"filter_class": DurationFilter},
        # models.SmallIntegerField: {"filter_class": NumberFilter},
        # models.AutoField: {"filter_class": NumberFilter},
        # models.PositiveIntegerField: {"filter_class": NumberFilter},
        # models.PositiveSmallIntegerField: {"filter_class": NumberFilter},
        # models.UUIDField: {"filter_class": UUIDFilter},
        # Forward relationships
        models.OneToOneField: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ForeignKey: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "to_field_name": f.remote_field.field_name,
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        models.ManyToManyField: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        # Reverse relationships
        OneToOneRel: {
            "filter_class": fields.ModelChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
                "null_label": settings.NULL_CHOICE_LABEL if f.null else None,
            },
        },
        ManyToOneRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
        ManyToManyRel: {
            "filter_class": fields.ModelMultipleChoiceFilter,
            "extra": lambda f: {
                "queryset": remote_queryset(f),
                "endpoint": f.related_model.get_representation_endpoint(),
                "value_key": f.related_model.get_representation_value_key(),
                "label_key": f.related_model.get_representation_label_key(),
            },
        },
    }

    @classmethod
    def filter_class_for_remote_filter(cls):
        if hasattr(cls, "get_filter_class_for_remote_filter") and callable(cls.get_filter_class_for_remote_filter()):
            return cls.get_filter_class_for_remote_filter()
        return cls

    def __init__(self, data=None, queryset=None, *, request=None, prefix=None, view=None):
        super().__init__(data=data, queryset=queryset, request=request, prefix=prefix)
        remote_filters = add_filters.send(sender=self.filter_class_for_remote_filter(), request=request)
        self.view = view
        for _, res in remote_filters:
            if res:
                for remote_filter_key, remote_filter in res.items():
                    remote_filter.column_field_name = remote_filter_key
                    self.filters[remote_filter_key] = remote_filter

    @classmethod
    def filter_for_lookup(cls, field, lookup_type):
        if isinstance(field, GeneratedField):
            return cls.filter_for_lookup(field.output_field, lookup_type)
        if lookup_type == "exact" and getattr(field, "choices", None):
            filter_class, params = fields.ChoiceFilter, {"choices": field.choices}
        else:
            filter_class, params = super().filter_for_lookup(field, lookup_type)

        # Check if it is a decimal field:
        if hasattr(field, "decimal_places"):
            params["precision"] = field.decimal_places

        if hasattr(field, "verbose_name"):
            params["label"] = field.verbose_name

        return filter_class, params

    @classmethod
    def get_dependency_map(cls):
        if hasattr(cls, "Meta"):
            yield from getattr(cls.Meta, "dependency_map", dict()).items()

    @classmethod
    def get_filters(cls):
        filters = dict(super().get_filters())
        remote_filters = add_filters.send(sender=cls.filter_class_for_remote_filter())
        for _, res in remote_filters:
            if res:
                for remote_filter_key, remote_filter in res.items():
                    remote_filter.column_field_name = remote_filter_key
                    filters[remote_filter_key] = remote_filter

        for field, help_text in getattr(cls, "help_texts", {}).items():
            filters[field].help_text = help_text

        # Get dependency map specified in Meta and append the argument accordingly
        for field, values in cls.get_dependency_map():
            for value in values:
                filters[field].depends_on.append({"field": value, "options": {}})

        excluding_fields = {}
        for name, field in filters.items():
            # if allow_exclude is true, we add a copy of the field with the parameter exclude=True
            # (to use `exclude` queryset method instead of `filter`) and add this with the suffix __{cls.DEFAULT_EXCLUDE_FILTER_LOOKUP}
            if field.allow_exclude:
                excluding_field = copy(field)
                excluding_field.exclude = True
                excluding_field.excluded_filter = True
                excluding_field.hidden = True
                excluding_field.required = False
                excluding_fields[f"{name}__{cls.DEFAULT_EXCLUDE_FILTER_LOOKUP}"] = excluding_field
        filters.update(excluding_fields)
        return OrderedDict(filters)

    def extract_required_field_labels(self):
        return [label for label, filter in self.base_filters.items() if getattr(filter, "required", False)]

    def filter_queryset(self, queryset):
        queryset = super().filter_queryset(queryset)
        check_required_filters(self.request, self.view, self)
        return queryset

    def fake_filter(self, queryset, name, value):
        return queryset

    # we enable for all filterset a way to easily narrow down the result based on related name not null
    notnull_related_name = fields.CharFilter(hidden=True, method="filter_notnull_related_name")

    def filter_notnull_related_name(self, queryset, name, value):
        if value:
            with suppress(FieldError):
                return queryset.filter(**{f"{value}__isnull": False}).distinct()
        return queryset



# Generated by Django 5.0.9 on 2024-12-05 13:27

import django.utils.timezone
from django.db import migrations, models


def set_created(apps, schema_editor):
    Notification = apps.get_model("notifications", "Notification")
    Notification.objects.filter(sent__isnull=False).update(created=models.F("sent"))


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0005_alter_notification_endpoint"),
    ]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="created",
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.RunPython(set_created),
    ]



# Generated by Django 4.1 on 2022-12-16 14:16

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("agenda", "0004_alter_calendaritem_item_type"),
    ]

    operations = [
        migrations.CreateModel(
            name="Building",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                (
                    "address",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="conference_room_building",
                        to="directory.addresscontact",
                        verbose_name="Address",
                    ),
                ),
            ],
            options={
                "verbose_name": "Building",
            },
        ),
        migrations.AddConstraint(
            model_name="building",
            constraint=models.UniqueConstraint(
                fields=("name", "address"),
                name="unique_building",
            ),
        ),
        migrations.CreateModel(
            name="ConferenceRoom",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                ("email", models.EmailField(verbose_name="Email Address", unique=True)),
                ("capacity", models.PositiveIntegerField(blank=True, null=True, verbose_name="Capacity")),
                (
                    "is_videoconference_capable",
                    models.BooleanField(default=True, verbose_name="Capable of Videoconferencing"),
                ),
                (
                    "building",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="conference_room",
                        to="agenda.building",
                        verbose_name="Building",
                    ),
                ),
            ],
            options={
                "verbose_name": "Conference Room",
            },
        ),
        migrations.AddConstraint(
            model_name="conferenceroom",
            constraint=models.UniqueConstraint(
                fields=("name", "building"),
                name="unique_conference_room",
            ),
        ),
        migrations.AddField(
            model_name="calendaritem",
            name="conference_room_new",
            field=models.ForeignKey(
                related_name="calendar_item",
                to="agenda.conferenceroom",
                verbose_name="Conference Room",
                on_delete=models.SET_NULL,
                null=True,
                blank=True,
            ),
        ),
        migrations.RemoveField(
            model_name="calendaritem",
            name="conference_room",
        ),
        migrations.RenameField(
            model_name="calendaritem",
            old_name="conference_room_new",
            new_name="conference_room",
        ),
    ]



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse

from wbcore import serializers as wb_serializers

from ..models import Company, CustomerStatus, Person
from .entries import (
    CompanyTypeRepresentationSerializer,
    CustomerStatusRepresentationSerializer,
    EntryModelSerializer,
    PersonRepresentationSerializer,
)


class CompanyModelSerializer(EntryModelSerializer):
    customer_status = wb_serializers.PrimaryKeyRelatedField(
        queryset=CustomerStatus.objects.all(),
        label=_("Customer Status"),
    )
    _customer_status = CustomerStatusRepresentationSerializer(source="customer_status")
    employees = wb_serializers.PrimaryKeyRelatedField(
        queryset=Person.objects.all(),
        default=wb_serializers.DefaultFromGET("employees", many=True),
        required=False,
        many=True,
    )
    _employees = PersonRepresentationSerializer(source="employees", many=True)
    is_primary_employer = wb_serializers.BooleanField(read_only=True)
    tier = wb_serializers.CharField(help_text=settings.DEFAULT_TIERING_HELP_TEXT, label=_("Tier"), required=False)
    _type = CompanyTypeRepresentationSerializer(source="type")

    @wb_serializers.register_resource()
    def employees_list(self, instance, request, user):
        return {"employees": reverse("wbcore:directory:employer-employee-list", args=[instance.id], request=request)}

    class Meta:
        model = Company
        read_only_fields = (
            "activity_heat",
            "entry_type",
        )
        fields = (
            "id",
            "computed_str",
            "activity_heat",
            "addresses",
            "cities",
            "_cities",
            "customer_status",
            "_customer_status",
            "description",
            "employees",
            "_employees",
            "entry_type",
            "headcount",
            "is_draft_entry",
            "is_primary_employer",
            "name",
            "primary_address",
            "primary_email",
            "primary_manager_repr",
            "primary_telephone",
            "primary_website",
            "primary_social",
            "profile_image",
            "salutation",
            "signature",
            "tier",
            "type",
            "_type",
            "_additional_resources",
            "primary_manager",
            "_primary_manager",
        )


class CompanyModelListSerializer(CompanyModelSerializer):
    eer_id = wb_serializers.CharField(default="", required=False, read_only=True)
    is_primary_employer = wb_serializers.BooleanField(read_only=True)

    @wb_serializers.register_resource()
    def delete(self, instance, request, user):
        if "view" in request.parser_context:
            try:
                person_id = request.parser_context["view"].kwargs.get("person_id", None)
            except AttributeError:
                return {}
            if person_id and instance.eer_id:
                if not (instance.is_primary_employer is True and len(self.instance) > 1):
                    return {
                        "delete": reverse(
                            "wbcore:directory:employeremployee-delete",
                            args=[instance.eer_id],
                            request=request,
                        )
                    }
        return {}

    @wb_serializers.register_resource()
    def eer_relationship_instance(self, instance, request, user):
        if "view" in request.parser_context:
            try:
                person_id = request.parser_context["view"].kwargs.get("person_id", None)
            except AttributeError:
                return {}
            if person_id and instance.eer_id:
                return {
                    "eer_relationship_instance": reverse(
                        "wbcore:directory:employeremployee-detail",
                        args=[instance.eer_id],
                        request=request,
                    )
                }
        return {}

    class Meta:
        model = Company

        fields = (
            "id",
            "name",
            "activity_heat",
            "cities",
            "_cities",
            "customer_status",
            "_customer_status",
            "eer_id",
            "is_primary_employer",
            "primary_address",
            "primary_email",
            "primary_manager_repr",
            "primary_telephone",
            "primary_website",
            "primary_social",
            "last_event",
            "last_event_period_endswith",
            "tier",
            "type",
            "_type",
            "_additional_resources",
        )


class BankModelSerializer(wb_serializers.ModelSerializer):
    primary_address = wb_serializers.CharField(
        allow_null=True, label=_("Primary Address"), read_only=True, required=False
    )
    primary_email = wb_serializers.CharField(allow_null=True, label=_("Primary Email"), required=False, read_only=True)
    primary_telephone = wb_serializers.TelephoneField(
        allow_null=True, label=_("Primary Telephone"), read_only=True, required=False
    )
    _relationship_managers = PersonRepresentationSerializer(many=True, source="relationship_managers")

    def create(self, validated_data: dict) -> Company:
        validated_data.pop("primary_manager", None)
        return super().create(validated_data)

    class Meta:
        model = Company
        fields = (
            "id",
            "name",
            "primary_address",
            "primary_email",
            "primary_telephone",
            "_relationship_managers",
            "relationship_managers",
        )



import mimetypes
import pathlib
import uuid

from django.db import models
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _


def upload_to(instance, filename):
    return f"markdown/assets/{instance.filename}"


class Asset(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    file = models.FileField(max_length=256, upload_to=upload_to)
    content_type = models.CharField(max_length=32, null=True, blank=True)
    file_url_name = models.CharField(max_length=1024, null=True, blank=True)

    # public = models.BooleanField(default=True)
    class Meta:
        verbose_name = _("Asset")
        verbose_name_plural = _("Assets")
        db_table = "bridger_asset"

    def __str__(self) -> str:
        return str(self.id)

    @property
    def filename(self):
        if suffix := pathlib.Path(self.file.name).suffix:
            return f"{self.id}{suffix}"
        return self.id


@receiver(models.signals.pre_save, sender="wbcore.Asset")
def generate_content_type(sender, instance, **kwargs):
    content_type, encoding = mimetypes.guess_type(pathlib.Path(instance.file.name))
    instance.content_type = content_type
    instance.file_url_name = instance.filename



# Generated by Django 4.1.8 on 2023-04-14 13:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("directory", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="entry",
            name="last_event",
            field=models.ForeignKey(
                blank=True,
                help_text="The last event of this entry. Can be ordered by end date.",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="last_event_of",
                to="agenda.calendaritem",
                verbose_name="Last Event",
            ),
        ),
        migrations.RemoveField(
            model_name="company",
            name="asset_allocation",
        ),
        migrations.RemoveField(
            model_name="company",
            name="assets_under_management",
        ),
        migrations.RemoveField(
            model_name="company",
            name="discretionairy_advisory",
        ),
        migrations.RemoveField(
            model_name="company",
            name="geographic_focus",
        ),
    ]



from .contacts import (
    AddressContactEntryTitleConfig,
    AddressContactTitleConfig,
    BankingContactEntryTitleConfig,
    BankingContactTitleConfig,
    EmailContactEntryTitleConfig,
    EmailContactTitleConfig,
    SocialMediaContactEntryTitleConfig,
    TelephoneContactEntryTitleConfig,
    TelephoneContactTitleConfig,
    WebsiteContactEntryTitleConfig,
)
from .entries import (
    CompanyModelTitleConfig,
    PersonModelTitleConfig,
    UserIsManagerTitleConfig,
)
from .relationships import (
    ClientManagerTitleConfig,
    EmployerEmployeeRelationshipTitleConfig,
    RelationshipEntryModelTitleConfig,
    RelationshipTypeModelTitleConfig,
    UserIsClientTitleConfig,
)
from .utils import (
    CompanyTypeTitleConfig,
    CustomerStatusTitleConfig,
    PositionTitleConfig,
    SpecializationTitleConfig,
)



from abc import ABC, abstractclassmethod

from wbcore.contrib.notifications.models import Notification


class AbstractNotificationBackend(ABC):
    @abstractclassmethod
    def send_notification(cls, notification: Notification): ...

    @abstractclassmethod
    def get_configuration(cls) -> dict: ...



# Generated by Django 4.2.6 on 2023-11-16 10:08

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("wbcore", "0012_delete_notification"),
    ]
    database_operations = [migrations.AlterModelTable("ColorGradient", "color_colorgradient")]

    state_operations = [migrations.DeleteModel("ColorGradient")]

    operations = state_operations



from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.shortcuts import render


def embedded_view_example(request: HttpRequest) -> HttpResponse:
    return render(request, "example_app/embedded_view.html", {"version": settings.FRONTEND_VERSION})



from rest_framework.reverse import reverse

from wbcore import serializers
from wbcore.contrib.notifications.models import (
    NotificationType,
    NotificationTypeSetting,
)


class NotificationTypeRepresentationSerializer(serializers.RepresentationSerializer):
    class Meta:
        model = NotificationType
        fields = (
            "id",
            "code",
            "title",
            "help_text",
        )


class NotificationTypeSettingModelSerializer(serializers.ModelSerializer):
    _notification_type = NotificationTypeRepresentationSerializer(source="notification_type")
    help_text = serializers.CharField()
    locked_icon = serializers.IconSelectField(read_only=True)
    locked = serializers.BooleanField(read_only=True)
    _update_url = serializers.SerializerMethodField()

    def get__update_url(self, obj):
        if not obj.notification_type.is_lock:
            return reverse(
                "wbcore:notifications:notification_type_setting-list", args=[], request=self.context["request"]
            )
        return None

    class Meta:
        read_only_fields = ("user", "notification_type", "help_text")
        model = NotificationTypeSetting
        fields = (
            "id",
            "notification_type",
            "_notification_type",
            "help_text",
            "user",
            "enable_web",
            "enable_mobile",
            "enable_email",
            "locked",
            "locked_icon",
            "_additional_resources",
            "_update_url",
        )



import pytest
from pytest import FixtureRequest
from rest_framework import status
from rest_framework.test import APIRequestFactory

from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.io.factories import (
    DataBackendFactory,
    ImportSourceFactory,
    ParserHandlerFactory,
    SourceFactory,
)
from wbcore.contrib.io.models import DataBackend, ImportSource, ParserHandler
from wbcore.contrib.io.serializers import (
    ImportSourceModelSerializer,
    ParserHandlerModelSerializer,
)
from wbcore.contrib.io.viewsets import (
    DataBackendRepresentationViewSet,
    ImportSourceModelViewSet,
    ImportSourceRepresentationViewSet,
    ParserHandlerModelViewSet,
    ParserHandlerRepresentationViewSet,
    SourceModelViewSet,
    SourceRepresentationViewSet,
)

BATCH_SIZE = 3


@pytest.mark.django_db
@pytest.mark.with_db
class TestViewsets:
    @pytest.fixture
    def request_factory(self):
        return APIRequestFactory()

    @pytest.fixture
    def super_user(self):
        return SuperUserFactory()

    @pytest.fixture
    def import_sources(self):
        return ImportSourceFactory.create_batch(BATCH_SIZE)

    @pytest.fixture
    def import_source(self, import_sources):
        return import_sources[0]

    @pytest.fixture
    def import_source_build(self):
        parser_handler = ParserHandlerFactory()
        source = SourceFactory()

        instance = ImportSourceFactory.build(parser_handler=None, source=None)

        serialized_data = ImportSourceModelSerializer(instance).data

        serialized_data["parser_handler"] = parser_handler.id
        serialized_data["source"] = source.id

        return serialized_data

    @pytest.fixture
    def sources(self):
        return SourceFactory.create_batch(BATCH_SIZE)

    @pytest.fixture
    def source(self, sources):
        return sources[0]

    @pytest.fixture
    def parser_handlers(self):
        return ParserHandlerFactory.create_batch(BATCH_SIZE)

    @pytest.fixture
    def parser_handler(self, parser_handlers):
        return parser_handlers[0]

    @pytest.fixture
    def parser_handler_build(self):
        instance = ParserHandlerFactory.build()
        return ParserHandlerModelSerializer(instance).data

    @pytest.fixture
    def data_backends(self):
        DataBackend.objects.all().delete()
        return DataBackendFactory.create_batch(BATCH_SIZE)

    @pytest.fixture
    def data_backend(self, data_backends):
        return data_backends[0]

    @pytest.mark.parametrize(
        "vs, batch",
        [
            (ImportSourceRepresentationViewSet, "import_sources"),
            (ImportSourceModelViewSet, "import_sources"),
            (SourceRepresentationViewSet, "sources"),
            (SourceModelViewSet, "sources"),
            (ParserHandlerRepresentationViewSet, "parser_handlers"),
            (ParserHandlerModelViewSet, "parser_handlers"),
            (DataBackendRepresentationViewSet, "data_backends"),
        ],
    )
    def test_get(self, request_factory, super_user, vs, batch, request: FixtureRequest):
        # Arrange
        request.getfixturevalue(batch)
        get_request = request_factory.get("")
        get_request.user = super_user
        viewset = vs.as_view({"get": "list"})
        # Act
        response = viewset(get_request)
        # Assert
        assert len(response.data["results"]) == BATCH_SIZE
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "vs, instance",
        [
            (ImportSourceRepresentationViewSet, "import_source"),
            (ImportSourceModelViewSet, "import_source"),
            (SourceRepresentationViewSet, "source"),
            (SourceModelViewSet, "source"),
            (ParserHandlerRepresentationViewSet, "parser_handler"),
            (ParserHandlerModelViewSet, "parser_handler"),
            (DataBackendRepresentationViewSet, "data_backend"),
        ],
    )
    def test_retrieve(self, request_factory, super_user, vs, instance, request: FixtureRequest):
        # Arrange
        entry = request.getfixturevalue(instance)
        get_request = request_factory.get("")
        get_request.user = super_user
        viewset = vs.as_view({"get": "retrieve"})
        # We need to work with lookup_field since SourceModelViewSet uses uuid as lookup field
        lookup_field = getattr(vs, "lookup_field", "pk")
        lookup_value = getattr(entry, lookup_field)
        # Act
        response = viewset(get_request, **{lookup_field: lookup_value})
        lookup_field = "id" if lookup_field != "uuid" else "uuid"
        instance = response.data.get("instance")
        # Assert
        assert instance is not None
        assert str(instance[lookup_field]) == str(lookup_value)
        assert not response.data.get("results")
        assert response.status_code == status.HTTP_200_OK

    @pytest.mark.parametrize(
        "vs, data, creatable",
        [
            (ImportSourceModelViewSet, "import_source_build", True),
            (ParserHandlerModelViewSet, "parser_handler_build", False),
        ],
    )
    def test_create(self, request_factory, super_user, vs, data, creatable, request: FixtureRequest):
        # Arrange
        build_data = request.getfixturevalue(data)
        post_request = request_factory.post("", data=build_data, format="json")
        post_request.user = super_user
        viewset = vs.as_view({"post": "create"})
        # Act
        response = viewset(post_request)
        expected_status_code = status.HTTP_201_CREATED if creatable else status.HTTP_405_METHOD_NOT_ALLOWED
        # Assert
        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "vs, batch, model, deletable",
        [
            (ImportSourceModelViewSet, "import_sources", ImportSource, True),
            (ParserHandlerModelViewSet, "parser_handlers", ParserHandler, False),
        ],
    )
    def test_delete(
        self,
        request_factory,
        super_user,
        vs,
        batch,
        model,
        deletable,
        request: FixtureRequest,
    ):
        # Arrange
        entries = request.getfixturevalue(batch)
        entry_id = entries[0].id
        delete_request = request_factory.delete("", args=entry_id)
        delete_request.user = super_user
        viewset = vs.as_view({"delete": "destroy"})
        # Act
        response = viewset(delete_request, pk=entry_id)
        # Assert
        if deletable:
            assert response.status_code == status.HTTP_204_NO_CONTENT
            assert model.objects.count() == BATCH_SIZE - 1
            assert not model.objects.filter(id=entry_id).exists()
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            assert model.objects.count() == BATCH_SIZE
            assert model.objects.filter(id=entry_id).exists()

    @pytest.mark.parametrize(
        "vs, entry, new_data, updatable",
        [
            (ImportSourceModelViewSet, "import_source", "import_source_build", True),
            (
                ParserHandlerModelViewSet,
                "parser_handler",
                "parser_handler_build",
                False,
            ),
        ],
    )
    def test_put(
        self,
        request_factory,
        super_user,
        vs,
        entry,
        new_data,
        updatable,
        request: FixtureRequest,
    ):
        # Arrange
        entry = request.getfixturevalue(entry)
        new_data = request.getfixturevalue(new_data)
        new_data["id"] = entry.id
        put_request = request_factory.put("", data=new_data, format="json")
        put_request.user = super_user
        viewset = vs.as_view({"put": "update"})
        # Act
        response = viewset(put_request, pk=entry.id)
        expected_status_code = status.HTTP_200_OK if updatable else status.HTTP_405_METHOD_NOT_ALLOWED
        # Assert
        assert response.status_code == expected_status_code

    @pytest.mark.parametrize(
        "vs, entry, field, updatable",
        [
            (ImportSourceModelViewSet, "import_source", "origin", True),
            (ParserHandlerModelViewSet, "parser_handler", "parser", False),
        ],
    )
    def test_patch(
        self,
        request_factory,
        super_user,
        vs,
        entry,
        field,
        updatable,
        request: FixtureRequest,
    ):
        # Arrange
        instance = request.getfixturevalue(entry)
        old_field_data = getattr(instance, field)
        new_field_data = "Foo Bar"
        patch_request = request_factory.patch("", data={field: new_field_data})
        patch_request.user = super_user
        viewset = vs.as_view({"patch": "partial_update"})
        # Act
        response = viewset(patch_request, pk=instance.id)
        instance.refresh_from_db()
        # Assert
        if updatable:
            assert response.status_code == status.HTTP_200_OK
            assert getattr(instance, field) == new_field_data
        else:
            assert response.status_code == status.HTTP_405_METHOD_NOT_ALLOWED
            assert getattr(instance, field) == old_field_data



from .contacts import (
    AddressContactFactory,
    BankingContactFactory,
    EmailContactFactory,
    TelephoneContactFactory,
    WebsiteContactFactory,
    SocialMediaContactFactory,
)
from .entries import (
    BankCompanyTypeFactory,
    BankFactory,
    ClientFactory,
    CompanyFactory,
    CompanyTypeFactory,
    CompanyWithEmployerEmployeeRelationshipFactory,
    CustomerStatusFactory,
    EmployersCompanyFactory,
    EntryFactory,
    PersonFactory,
    PersonWithEmployerFactory,
    RandomClientFactory,
    SpecializationFactory,
    UnemployedPersonFactory,
    PersonSignatureFactory,
)
from .relationships import (
    ClientManagerRelationshipFactory,
    EmployerEmployeeRelationshipFactory,
    ParentRelationshipTypeFactory,
    PositionFactory,
    RelationshipFactory,
    RelationshipTypeFactory,
    UserIsClientPersonFactory,
    UserIsManagerEntryFactory,
)



import os

from django_filters.constants import EMPTY_VALUES

from .defaults import RequiredFilterMissing


def check_required_filters(request, view, filter):
    if "PYTEST_CURRENT_TEST" not in os.environ and getattr(view, "action", "list") == "list":
        errors = []
        for required_field in [
            label for label, filter in filter.base_filters.items() if getattr(filter, "required", False)
        ]:
            if filter.form.cleaned_data.get(required_field, None) in EMPTY_VALUES:
                errors.append(required_field)
        if errors:
            raise RequiredFilterMissing(
                detail={
                    "filter_errors": {field: ["This filter is required. Please specify a value"] for field in errors},
                }
            )



import pytest
from django.forms.models import model_to_dict
from rest_framework.exceptions import ValidationError
from rest_framework.test import APITestCase
from wbcore.contrib.example_app.factories import (
    EventFactory,
    EventTypeFactory,
    LeagueFactory,
    MatchFactory,
    SportFactory,
    SportPersonFactory,
    StadiumFactory,
    TeamFactory,
)
from wbcore.contrib.example_app.serializers import (
    EventErrorMessages,
    EventModelSerializer,
    EventTypeModelSerializer,
    LeaguePlayerStatisticsModelSerializer,
    LeagueTeamStatisticsModelSerializer,
    MatchErrorMessages,
    MatchModelSerializer,
)


@pytest.mark.django_db
class TestMatchModelSerializer(APITestCase):
    def test_match_serializer(self):
        match_data: dict = model_to_dict(
            MatchFactory.build(away=TeamFactory(), home=TeamFactory(), sport=SportFactory(), stadium=StadiumFactory())
        )
        match_serializer = MatchModelSerializer(data=match_data)
        self.assertTrue(match_serializer.is_valid())

    def test_match_same_teams(self):
        team = TeamFactory()
        match_data: dict = model_to_dict(
            MatchFactory.build(away=team, home=team, sport=SportFactory(), stadium=StadiumFactory())
        )
        with self.assertRaisesMessage(ValidationError, MatchErrorMessages.same_teams.value):
            league_serializer = MatchModelSerializer(data=match_data)
            self.assertFalse(league_serializer.is_valid(raise_exception=True))

    def test_match_same_date(self):
        match = MatchFactory()
        match_data: dict = model_to_dict(match)
        with pytest.raises(ValidationError):
            MatchModelSerializer(data=match_data).is_valid(raise_exception=True)

    def test_match_wrong_sport(self):
        match_data: dict = model_to_dict(
            MatchFactory.build(
                away=TeamFactory(),
                home=TeamFactory(),
                stadium=StadiumFactory(),
                sport=SportFactory(),
                league=LeagueFactory(),
            )
        )
        with self.assertRaisesMessage(ValidationError, MatchErrorMessages.wrong_sport.value):
            league_serializer = MatchModelSerializer(data=match_data)
            self.assertFalse(league_serializer.is_valid(raise_exception=True))


@pytest.mark.django_db
class TestEventTypeModelSerializer(APITestCase):
    def test_event_type_serializer(self):
        event_type_data: dict = model_to_dict(EventTypeFactory.build(sport=SportFactory()))
        event_type_serializer = EventTypeModelSerializer(data=event_type_data)
        self.assertTrue(event_type_serializer.is_valid())

    def test_event_type_same_name(self):
        event_type = EventTypeFactory()
        event_type_data: dict = model_to_dict(EventTypeFactory.build(name=event_type.name, sport=event_type.sport))
        with pytest.raises(ValidationError):
            EventTypeModelSerializer(data=event_type_data).is_valid(raise_exception=True)


@pytest.mark.django_db
class TestEventModelSerializer(APITestCase):
    def test_event_serializer(self):
        match = MatchFactory()
        event_type = EventTypeFactory(sport=match.sport)
        event_data: dict = model_to_dict(
            EventFactory.build(person=SportPersonFactory(), event_type=event_type, match=match)
        )
        event_serializer = EventModelSerializer(data=event_data)
        self.assertTrue(event_serializer.is_valid())

    def test_event_wrong_duration(self):
        sport = SportFactory(match_duration=60)
        event_type = EventTypeFactory(sport=sport)
        match = MatchFactory(sport=sport)
        event: dict = model_to_dict(
            EventFactory.build(person=SportPersonFactory(), minute=61, event_type=event_type, match=match)
        )
        with self.assertRaisesMessage(ValidationError, EventErrorMessages.wrong_duration.value):
            event_serializer = EventModelSerializer(data=event)
            self.assertFalse(event_serializer.is_valid(raise_exception=True))

    def test_event_duplication(self):
        event = EventFactory()
        event_data = model_to_dict(event)

        with pytest.raises(ValidationError):
            EventModelSerializer(data=event_data).is_valid(raise_exception=True)

    def test_event_wrong_type(self):
        event_type = EventTypeFactory()
        match = MatchFactory()
        event_data: dict = model_to_dict(
            EventFactory.build(person=SportPersonFactory(), event_type=event_type, match=match)
        )
        with self.assertRaisesMessage(
            ValidationError, EventErrorMessages.wrong_event_type.value.format(match.sport.name)
        ):
            event_serializer = EventModelSerializer(data=event_data)
            self.assertFalse(event_serializer.is_valid(raise_exception=True))


@pytest.mark.django_db
class TestLeaguePlayerStatisticsModelSerializer(APITestCase):
    def test_league_player_statistics_serializer(self):
        statistic_data = {"person_id": 1, "person_name": "Foo", "count": 1, "id": "id"}
        statistic_serializer = LeaguePlayerStatisticsModelSerializer(data=statistic_data)
        self.assertTrue(statistic_serializer.is_valid())


@pytest.mark.django_db
class TestLeagueTeamStatisticsModelSerializer(APITestCase):
    def test_league_player_statistics_serializer(self):
        statistic_data = {"team_id": 1, "team_name": "Foo", "count": 1, "id": "id"}
        statistic_serializer = LeagueTeamStatisticsModelSerializer(data=statistic_data)
        self.assertTrue(statistic_serializer.is_valid())



import typing
from dataclasses import dataclass

from wbcore.metadata.utils import prefix_key

if typing.TYPE_CHECKING:
    from ..display import Display, SerializedDisplay


type SerializedInline = dict[str, str | SerializedDisplay | list[str] | bool]


@dataclass
class Inline:
    """An inline can be used to display a different endpoint in a form

    When generating a form, only fields of the original endpoint can be displayed. An inline allows you to specify
    remote content, for example a chart or a table, which the form asynchronously loads. This is very useful to
    create dashboard like views and to seperate the load to multiple endpoints

    Attributes:
        key: A reference for the positioning on the grid
        endpoint: A reference to the endpoint in the additional resources
        display: An optional `Display` to override the `display` behind the `endpoint` in the additional resources

    """

    key: str
    endpoint: str
    title: str | None = None
    display: typing.Optional["Display"] = None
    aligned_keys: list[str] | None = None
    hide_controls: bool = False

    def serialize(self, key_prefix: str | None = None) -> SerializedInline:
        """Serializes this `Inline`
        Attributes:
            key_prefix(Optional): if specified, append a prefix to the inline key
        Returns:
            A dictionairy containing all serialized fields

        """
        serialized_inline: SerializedInline = {
            "key": self.key,
            "endpoint": prefix_key(self.endpoint, key_prefix),
            "hide_controls": self.hide_controls,
        }

        if self.title is not None:
            serialized_inline["title"] = self.title

        if self.display:
            serialized_inline["display"] = self.display.serialize()

        if self.aligned_keys:
            serialized_inline["aligned_keys"] = self.aligned_keys

        return serialized_inline



from rest_framework import serializers

from .list import ListFieldMixin
from .mixins import WBCoreSerializerFieldMixin
from .types import WBCoreType


class ChoiceMixin:
    choices: dict[str, str]

    def __init__(self, *args, group_key_mapping: dict[str, str] | None = None, **kwargs):
        """

        Args:
            group_key_mapping (dict[str, str] | None, optional): An optional mapping that provides
                a group for every choice value to group the choices in the drop down by. Defaults to None.
        """
        self.group_key_mapping = group_key_mapping or dict()
        super().__init__(*args, **kwargs)

    def _get_choices_representation(self):
        return [
            {"group": self.group_key_mapping.get(value), "value": value, "label": label}
            for value, label in self.choices.items()
        ]


class ChoiceField(WBCoreSerializerFieldMixin, ChoiceMixin, serializers.ChoiceField):
    field_type = WBCoreType.SELECT.value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["choices"] = self._get_choices_representation()
        return key, representation


class MultipleChoiceField(ListFieldMixin, WBCoreSerializerFieldMixin, ChoiceMixin, serializers.MultipleChoiceField):
    field_type = WBCoreType.SELECT.value

    def get_representation(self, request, field_name) -> tuple[str, dict]:
        key, representation = super().get_representation(request, field_name)
        representation["multiple"] = True
        representation["choices"] = self._get_choices_representation()
        return key, representation

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        if isinstance(data, set):
            data = list(data)
        return data

    def to_representation(self, data):
        data = super().to_representation(data)
        if isinstance(data, set):
            data = list(data)
        return data


class LanguageChoiceField(ChoiceField):
    field_type = WBCoreType.LANGUAGE.value



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import SportFactory
from wbcore.contrib.example_app.models import Sport
from wbcore.contrib.example_app.serializers import SportModelSerializer
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import SportModelViewSet


@pytest.mark.django_db
class TestSportModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:sport-list")
        self.detail_url_str = "example_app:sport-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        sport = SportFactory()
        response = get_create_view(self.client, sport, self.user, self.list_url, SportModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(Sport.objects.filter(name=sport.name).exists())

    def test_detail_view(self):
        instance = SportFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_update_view(self):
        instance = SportFactory()
        instance.name = "Updated Instance"
        response = get_update_view(self.client, instance, SportModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_partial_update_view(self):
        instance = SportFactory()
        response = get_partial_view(self.client, instance.id, {"name": "Patched name"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_delete_view(self):
        instance = SportFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(Sport.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        first_name, second_name, third_name = "Team A", "Team B", "Team C"
        SportFactory(name=second_name)
        SportFactory(name=first_name)
        SportFactory(name=third_name)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], Sport.objects.count())
        self.assertEqual(response.data["results"][0]["name"], first_name)
        self.assertEqual(response.data["results"][1]["name"], second_name)
        self.assertEqual(response.data["results"][2]["name"], third_name)



# -*- encoding: utf-8 -*-

from django.conf import settings
from django.contrib import admin, messages
from django.contrib.admin.options import IS_POPUP_VAR
from django.contrib.admin.utils import unquote
from django.contrib.auth import update_session_auth_hash
from django.contrib.auth.forms import (
    AdminPasswordChangeForm,
    UserChangeForm,
    UserCreationForm,
)
from django.contrib.auth.models import Permission
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import re_path, reverse
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.html import escape
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.debug import sensitive_post_parameters

from .models import Token, User, UserActivity

csrf_protect_m = method_decorator(csrf_protect)
sensitive_post_parameters_m = method_decorator(sensitive_post_parameters())


@admin.register(Permission)
class PermissionAdmin(admin.ModelAdmin):
    list_display = ("name", "content_type", "codename")
    search_fields = ("name", "codename")


class TokenInline(admin.TabularInline):
    model = Token
    fields = (
        "key",
        "valid_until",
        "protected_view_name",
        "number_usage_left",
        "is_valid",
    )
    readonly_fields = ("key", "user")
    extra = 0
    raw_id_fields = ["user"]
    show_change_link = True


admin.site.register(UserActivity)


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):
    search_fields = ("key", "user__username", "user__email")

    list_display = (
        "user",
        "key",
        "valid_until",
        "protected_view_name",
        "number_usage_left",
        "is_valid",
    )


@admin.register(User)
class UserAdmin(admin.ModelAdmin):
    # add_form_template = 'admin/auth/user/add_form.html'
    change_user_password_template = None
    fieldsets = (
        (_("Main information"), {"fields": ("email", "username", "password", "profile")}),
        (_("Synchronization"), {"fields": ("metadata",)}),
        (
            _("Permissions"),
            {"fields": ("is_active", "is_register", "is_staff", "is_superuser", "groups", "user_permissions")},
        ),
    )
    raw_id_fields = ("profile",)
    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("email", "username", "password1", "password2"),
            },
        ),
    )
    form = UserChangeForm
    add_form = UserCreationForm
    change_password_form = AdminPasswordChangeForm

    list_display = ("email", "username", "uuid", "is_staff", "is_active", "is_register")
    list_filter = ("is_staff", "is_superuser", "is_active", "groups")
    search_fields = ("email", "username")
    ordering = ("email",)
    filter_horizontal = (
        "groups",
        "user_permissions",
    )

    def get_fieldsets(self, request, obj=None):
        if not obj:
            return self.add_fieldsets
        return super(UserAdmin, self).get_fieldsets(request, obj)

    def get_form(self, request, obj=None, **kwargs):
        """
        Use special form during user creation
        """
        defaults = {}
        if obj is None:
            defaults["form"] = self.add_form
        defaults.update(kwargs)
        return super(UserAdmin, self).get_form(request, obj, **defaults)

    def get_urls(self):
        return [
            re_path(
                r"^(.+)/password/$",
                self.admin_site.admin_view(self.user_change_password),
                name="auth_user_password_change",
            ),
        ] + super(UserAdmin, self).get_urls()

    def lookup_allowed(self, lookup, value):
        # See #20078: we don't want to allow any lookups involving passwords.
        if lookup.startswith("password"):
            return False
        return super(UserAdmin, self).lookup_allowed(lookup, value)

    @sensitive_post_parameters_m
    @csrf_protect_m
    @transaction.atomic
    def add_view(self, request, form_url="", extra_context=None):
        # It's an error for a user to have add permission but NOT change
        # permission for users. If we allowed such users to add users, they
        # could create superusers, which would mean they would essentially have
        # the permission to change users. To avoid the problem entirely, we
        # disallow users from adding users if they don't have change
        # permission.
        if not self.has_change_permission(request):
            if self.has_add_permission(request) and settings.DEBUG:
                # Raise Http404 in debug mode so that the user gets a helpful
                # error message.
                raise Http404(
                    _(
                        'Your user does not have the "Change user" permission. In '
                        "order to add users, Django requires that your user "
                        'account have both the "Add user" and "Change user" '
                        "permissions set."
                    )
                )
            raise PermissionDenied
        if extra_context is None:
            extra_context = {}
        username_field = self.model._meta.get_field(self.model.USERNAME_FIELD)
        defaults = {
            "auto_populated_fields": (),
            "username_help_text": username_field.help_text,
        }
        extra_context.update(defaults)
        return super(UserAdmin, self).add_view(request, form_url, extra_context)

    @sensitive_post_parameters_m
    def user_change_password(self, request, id, form_url=""):
        if not self.has_change_permission(request):
            raise PermissionDenied
        user = self.get_object(request, unquote(id))
        if user is None:
            raise Http404(
                _("%(name)s object with primary key %(key)r does not exist.")
                % {
                    "name": force_str(self.model._meta.verbose_name),
                    "key": escape(id),
                }
            )
        if request.method == "POST":
            form = self.change_password_form(user, request.POST)
            if form.is_valid():
                form.save()
                change_message = self.construct_change_message(request, form, None)
                self.log_change(request, user, change_message)
                msg = _("Password changed successfully.")
                messages.success(request, msg)
                update_session_auth_hash(request, form.user)
                return HttpResponseRedirect(
                    reverse(
                        "%s:%s_%s_change"
                        % (
                            self.admin_site.name,
                            user._meta.app_label,
                            user._meta.model_name,
                        ),
                        args=(user.pk,),
                    )
                )
        else:
            form = self.change_password_form(user)

        fieldsets = [(None, {"fields": list(form.base_fields)})]
        admin_form = admin.helpers.AdminForm(form, fieldsets, {})

        context = {
            "title": _("Change password: %s") % escape(user.get_username()),
            "adminForm": admin_form,
            "form_url": form_url,
            "form": form,
            "is_popup": (IS_POPUP_VAR in request.POST or IS_POPUP_VAR in request.GET),
            "add": True,
            "change": False,
            "has_delete_permission": False,
            "has_change_permission": True,
            "has_absolute_url": False,
            "opts": self.model._meta,
            "original": user,
            "save_as": False,
            "show_save": True,
        }
        context.update(self.admin_site.each_context(request))

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request,
            self.change_user_password_template or "admin/auth/user/change_password.html",
            context,
        )

    def response_add(self, request, obj, post_url_continue=None):
        """
        Determines the HttpResponse for the add_view stage. It mostly defers to
        its superclass implementation but is customized because the User model
        has a slightly different workflow.
        """
        # We should allow further modification of the user just added i.e. the
        # 'Save' button should behave like the 'Save and continue editing'
        # button except in two scenarios:
        # * The user has pressed the 'Save and add another' button
        # * We are adding a user in a popup
        if "_addanother" not in request.POST and IS_POPUP_VAR not in request.POST:
            request.POST = request.POST.copy()
            request.POST["_continue"] = 1
        return super(UserAdmin, self).response_add(request, obj, post_url_continue)



from unittest.mock import patch

import pandas as pd
import pytest
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.test.client import RequestFactory
from faker import Faker
from rest_framework.reverse import reverse
from rest_framework.test import force_authenticate

from ..enums import ExportFormat, get_django_import_export_format
from ..models import ExportSource, ParserHandler
from ..resources import ViewResource
from ..serializers import ParserHandlerModelSerializer
from ..viewsets import ParserHandlerModelViewSet
from .test_models import get_test_file

fake = Faker()


@pytest.mark.django_db
class TestExportSourceModel:
    def test_file_format(self, export_source):
        assert (
            export_source.file_format.CONTENT_TYPE
            == get_django_import_export_format(export_source.format).CONTENT_TYPE
        )

    def test_resource(self, export_source):
        r = export_source.resource
        assert isinstance(r, ViewResource)
        assert r.columns_map == export_source.resource_kwargs["columns_map"]
        assert r.serializer_class == ParserHandlerModelSerializer

    def test_empty_queryset(self, export_source):
        assert set(export_source.queryset) == set()

    def test_queryset(self, export_source_factory, parser_handler_factory):
        ph1 = parser_handler_factory.create()
        e1 = export_source_factory()
        assert set(e1.queryset) == {ph1}
        ph2 = parser_handler_factory.create()
        e2 = export_source_factory()

        assert set(e2.queryset) == {ph1, ph2}

    def test_queryset_with_filter(self, export_source_factory, parser_handler_factory):
        ph1 = parser_handler_factory.create()
        parser_handler_factory.create()  # noise

        # check with customer queryset and filtering
        query_str, query_params = ParserHandler.objects.filter(parser=ph1.parser).query.sql_with_params()
        export_source = export_source_factory.create(query_str=query_str, query_params=query_params)

        assert set(export_source.queryset) == {ph1}

    def test_get_export_filename(self, export_source):
        filename = export_source.get_export_filename()
        assert isinstance(filename, str)
        assert (
            export_source.file_format.get_extension() in filename
        )  # check that the filename is with the proper extension

    @patch("wbcore.contrib.io.models.send_notification")
    def test_notify(self, mock_fct, export_source):
        export_source.notify()
        assert mock_fct.call_count == 0  # 0 call count because there is neither no file nor status == processed

        export_source.status = ExportSource.Status.PROCESSED.value
        export_source.save()

        export_source.notify()
        assert mock_fct.call_count == 0  # Still need a file

        export_source.file = get_test_file()
        export_source.save()

        export_source.notify()
        assert mock_fct.call_count == 1  # Still need a file

    @patch.object(ExportSource, "notify")
    @pytest.mark.parametrize("export_source__format", [ExportFormat.CSV])
    def test_export_data(self, fct_mock, export_source, parser_handler):
        export_source.export_data()
        assert export_source.status == ExportSource.Status.PROCESSED.value
        with export_source.file.open() as file:
            df = pd.read_csv(file)
            assert df.loc[0, "id"] == parser_handler.id
            assert df.loc[0, "Parser"] == parser_handler.parser
            assert df.loc[0, "Handler"] == parser_handler.handler
        assert fct_mock.call_count == 1

    # INFO This test is obsolete
    # def test_view_export_under_pagination_limit(self, superuser, parser_handler_factory):
    #     parser_handler_factory.create()
    #
    #     url = reverse("wbcore:io:parserhandler-processexport", args=[])
    #     request = RequestFactory().get(url, {"export_format": 0})
    #     force_authenticate(request, user=superuser)
    #     view = ParserHandlerModelViewSet.as_view({"get": "process_export"})
    #     response = view(request)
    #     assert response.headers["Content-Type"] == "text/csv"

    def test_view_export_over_pagination_limit(self, superuser, parser_handler_factory):
        assert ExportSource.objects.count() == 0

        settings.WBIMPORT_EXPORT_DEFAULT_EXPORT_PAGINATION_LIMIT = 1

        parser_handler_factory.create()

        url = reverse("wbcore:io:parserhandler-processexport", args=[])
        request = RequestFactory().patch(url, {"format": 0}, content_type="application/json")
        request.query_params = {}

        force_authenticate(request, user=superuser)
        view = ParserHandlerModelViewSet.as_view({"patch": "process_export"})
        response = view(request)
        assert response.status_code == 200

        _view = ParserHandlerModelViewSet()
        _view.request = request
        query_str, query_params = _view.filter_queryset(_view.get_queryset()).query.sql_with_params()
        assert ExportSource.objects.get(
            creator=superuser,
            content_type=ContentType.objects.get_for_model(ParserHandler),
            resource_path="wbcore.contrib.io.resources.ViewResource",
            query_str=query_str,
            query_params=query_params,
        )



class Apps:
    _BASE_APPS = [
        "django.contrib.admin",
        "django.contrib.auth",
        "django.contrib.contenttypes",
        "django.contrib.sessions",
        "django.contrib.messages",
        "django.contrib.staticfiles",
        "django.contrib.sites",
        "rest_framework",
        "django_filters",
        "dynamic_preferences",
        "dynamic_preferences.users.apps.UserPreferencesConfig",
        "django_extensions",
        "corsheaders",
        "reversion",
        "reversion_compare",
        "guardian",
        "ordered_model",
        "wbcore",
        "wbcore.contrib.agenda",
        "wbcore.contrib.directory",
        "wbcore.contrib.authentication",
        "wbcore.contrib.notifications",
        "wbcore.contrib.io",
        "wbcore.contrib.currency",
        "wbcore.contrib.geography",
        "wbcore.contrib.tags",
        "wbcore.contrib.documents",
        "wbcore.contrib.workflow",
        "wbcore.contrib.color",
        "wbcore.contrib.guardian",
        "django_celery_beat",
        "modeltrans",
        "maintenance_mode",
    ]

    def _get_additional_and_wb_apps(self):
        return getattr(self, "ADDITIONAL_APPS", []) + getattr(self, "WB_ENDPOINTS", [])

    @property
    def INSTALLED_APPS(self):  # noqa
        return self._BASE_APPS + self._get_additional_and_wb_apps()


class DevApps(Apps):
    @property
    def INSTALLED_APPS(self):  # noqa
        apps = self._BASE_APPS

        if self.DEBUG:
            apps.append("debug_toolbar")

        apps.extend(self._get_additional_and_wb_apps())
        return apps



import pytest

from wbcore.contrib.pandas.fields import YearField as PandasYearField


class TestPandasYearField:
    @pytest.mark.parametrize(
        (
            "key",
            "label",
        ),
        [
            ("Foo", "Bar"),
        ],
    )
    def test_pandas_year_field_values(self, key, label):
        field = PandasYearField(key=key, label=label, precision=2)
        representation = field.to_dict()
        assert representation["precision"] == 0



import random
from typing import Iterator

from wbcore.contrib.dataloader.dataloaders import Dataloader

from .protocols import EntityTestDataProtocol


class RandomData(EntityTestDataProtocol, Dataloader):
    def data(self, n: int) -> Iterator[int]:
        for _ in self.entities:
            yield from map(lambda x: random.randint(0, 1000), range(n))


class RandomDataOver1000(EntityTestDataProtocol, Dataloader):
    def data(self, n: int) -> Iterator[int]:
        for _ in self.entities:
            yield from map(lambda x: random.randint(1000, 10000), range(n))



from configurations import values
from django.utils.translation import gettext_lazy as _
from markdown.extensions.tables import TableExtension
from markdown_blockdiag.extension import BlockdiagExtension

from wbcore.fsm.markdown_extensions import FSMExtension


class WBCore:
    CDN_BASE_ENDPOINT_URL = values.URLValue(
        "https://stainly-cdn.fra1.cdn.digitaloceanspaces.com/static", environ_prefix=None
    )

    FRONTEND_VERSION = values.Value("2.6.3", environ_prefix=None)
    FRONTEND_TEMPLATE = "wbcore/frontend.html"
    FRONTEND_MENU_CALENDAR = None

    BASE_ENDPOINT_URL = values.Value("", environ_prefix=None)

    BETA_BUTTON_VERSION = values.Value(None, environ_prefix=None)
    BETA_BUTTON_URL = values.URLValue(None, environ_prefix=None)
    BETA_BUTTON_TEXT = values.Value(_("Check out our upcoming beta version"), environ_prefix=None)

    WBCORE_MARKDOWN_TEMPLATE_TAGS = []
    WBCORE_MARKDOWN_EXTENSIONS = [
        TableExtension(),
        FSMExtension(),
        BlockdiagExtension(format="svg"),
    ]

    WBCORE_NOTIFICATION_TEMPLATE = values.Value("notifications/email_template.html", environ_prefix=None)
    WBCORE_NOTIFICATION_EMAIL_FROM = values.Value("no-reply@stainly-bench.com", environ_prefix=None)

    WBCORE_PLOTLY_CDN_URL = values.Value("https://cdn.plot.ly/plotly-latest.min.js", environ_prefix=None)
    WBCORE_EXTRA_CDN_URLS = values.ListValue([], environ_prefix=None)
    WBCORE_EXTRA_CSS_URLS = values.ListValue([], environ_prefix=None)
    WBCORE_PROFILE = values.Value("wbcore.contrib.authentication.configuration.resolve_profile", environ_prefix=None)
    WBCORE_DEFAULT_FRONTEND_USER_CONFIGURATION_ORDER = values.ListValue(["config__order"], environ_prefix=None)
    WBCORE_ADDITIONAL_DEFAULT_MODELVIEW_ATTRIBUTES = values.ListValue([], environ_prefix=None)
    WBCORE_NEW_NOTIFICATION_SYSTEM = values.BooleanValue(False, environ_prefix=None)

    DATA_UPLOAD_MAX_NUMBER_FIELDS = values.IntegerValue(2000, environ_prefix=None)
    WBCORE_IDLE_TIME = values.IntegerValue(30 * 60, environ_prefix=None)  # value in seconds
    WBCORE_IDLE_WARNING_TIME = values.IntegerValue(90, environ_prefix=None)  # value in seconds

    MESSAGE_STORAGE = "wbcore.messages.route_message_storage"

    @property
    def FRONTEND_CONTEXT(self):  # noqa
        base_url = f"{self.CDN_BASE_ENDPOINT_URL}/{self.FRONTEND_VERSION}/"
        return {
            "TITLE": "Workbench",
            "FONT_URL": "https://fonts.googleapis.com/css?family=Roboto:400,400i,500,500i,700,900&display=swap",
            "CDN_URLS": [self.WBCORE_PLOTLY_CDN_URL, *self.WBCORE_EXTRA_CDN_URLS],
            "CSS_URLS": self.WBCORE_EXTRA_CSS_URLS,
            "CONFIG_URL": "/wbcore/config/",
            "FAVICON_URL": f"{self.CDN_BASE_ENDPOINT_URL}/favicon.ico",
            "JS_URL": f"{base_url}main.js",
            "CSS_URL": f"{base_url}main.css",
            "JS_RUNTIME_URL": f"{base_url}runtime.js",
            "IDLE_TIME": self.WBCORE_IDLE_TIME * 1000,  # convert it into milliseconds
            "IDLE_WARNING_TIME": self.WBCORE_IDLE_WARNING_TIME * 1000,  # convert it into milliseconds
        }



from dataclasses import dataclass
from typing import TYPE_CHECKING

from .layouts.layouts import Layout, SerializedLayout

if TYPE_CHECKING:
    from .display import Display, SerializedDisplay


type SerializedPage = dict[str, None | str | dict[str, SerializedLayout] | SerializedDisplay]


@dataclass
class Page:
    """Represents a page, which may include an optional title and a dictionary defining layout boundaries along with a Layout. Alternatively, a list of subpages can also be specified."""

    title: str | None = None
    layouts: dict[str, Layout] | None = None
    display: "Display | None" = None

    def serialize(self, key_prefix: str | None = None, view_config=None, view=None, request=None) -> SerializedPage:
        """Serializes the `Page` into a dictionairy

        Returns:
            A dictionairy containing all fields of a `Page`

        """
        page: SerializedPage = {
            "title": self.title,
        }

        if self.layouts is not None:
            page["layouts"] = {key: layout.serialize(key_prefix=key_prefix) for key, layout in self.layouts.items()}

        elif self.display is not None:
            page["display"] = self.display.serialize(
                view_config=view_config, view=view, request=request, key_prefix=key_prefix, parent_page=self
            )

        return page



from django.utils.translation import gettext as _

from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import (
    Display,
    Inline,
    Layout,
    Page,
    Section,
    Style,
)
from wbcore.metadata.configs.display.instance_display.operators import default, lte
from wbcore.metadata.configs.display.list_display import ListDisplay
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class SportDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> ListDisplay:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="name", label=_("Name")),
                dp.Field(key="match_duration", label=_("Match Duration")),
                dp.Field(key="rules", label=_("Rules")),
            )
        )

    def get_instance_display(self) -> Display:
        leagues_section = Section(
            key="leagues_section",
            collapsible=False,
            title=_("Leagues"),
            display=Display(
                pages=[
                    Page(
                        title=_("Leagues"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["leagues_inline"]],
                                grid_template_columns=[
                                    "minmax(min-content, 1fr)",
                                ],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="leagues_inline", endpoint="existing_leagues")],
                            )
                        },
                    ),
                ]
            ),
        )
        event_types_section = Section(
            key="event_types_section",
            title=_("Event Types"),
            display=Display(
                pages=[
                    Page(
                        title=_("Event Types"),
                        layouts={
                            default(): Layout(
                                grid_template_areas=[["event_types_inline"]],
                                grid_template_columns=["minmax(min-content, 1fr)"],
                                grid_auto_rows=Style.MIN_CONTENT,
                                inlines=[Inline(key="event_types_inline", endpoint="associated_event_types")],
                            )
                        },
                    ),
                ]
            ),
        )

        is_create_display = "pk" not in self.view.kwargs

        return Display(
            pages=[
                Page(
                    layouts={
                        default(): Layout(
                            grid_template_areas=[
                                ["name", "leagues_section"],
                                ["match_duration", "leagues_section"],
                                ["rules", "leagues_section"],
                                ["event_types_section", "event_types_section"],
                            ]
                            if not is_create_display
                            else [["name", "match_duration", "rules"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(5),
                            sections=[leagues_section, event_types_section],
                        ),
                        lte(1000): Layout(
                            grid_template_areas=[
                                ["name", "match_duration"],
                                ["rules", "rules"],
                                ["leagues_section", "leagues_section"],
                                ["event_types_section", "event_types_section"],
                            ]
                            if not is_create_display
                            else [["name", "match_duration"], ["rules", "rules"]],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                            column_gap=Style.rem(3),
                            sections=[leagues_section, event_types_section],
                        ),
                        lte(650): Layout(
                            grid_template_areas=[
                                ["name"],
                                ["match_duration"],
                                ["rules"],
                            ],
                            grid_auto_columns="minmax(min-content, 1fr)",
                            grid_auto_rows=Style.MIN_CONTENT,
                        ),
                    },
                ),
            ]
        )



import pytest
from django.test import Client, TestCase
from django.urls import reverse
from wbcore.contrib.authentication.factories import SuperUserFactory
from wbcore.contrib.example_app.factories import StadiumFactory
from wbcore.contrib.example_app.models import Stadium
from wbcore.contrib.example_app.serializers import StadiumModelSerializer
from wbcore.contrib.example_app.tests.test_viewsets.test_utils_viewsets import (
    get_create_view,
    get_delete_view,
    get_detail_view,
    get_partial_view,
    get_update_view,
)
from wbcore.contrib.example_app.viewsets import StadiumModelViewSet


@pytest.mark.django_db
class TestStadiumModelViewSet(TestCase):
    def setUp(self):
        self.user = SuperUserFactory()
        self.client = Client()
        self.client.force_login(user=self.user)
        self.list_url = reverse("example_app:stadium-list")
        self.detail_url_str = "example_app:stadium-detail"

    def test_list_view(self):
        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)

    def test_create_view(self):
        stadium = StadiumFactory()
        response = get_create_view(self.client, stadium, self.user, self.list_url, StadiumModelViewSet)
        self.assertEqual(response.status_code, 201)
        self.assertTrue(Stadium.objects.filter(name=stadium.name).exists())

    def test_detail_view(self):
        instance = StadiumFactory()
        response = get_detail_view(self.client, instance.pk, self.detail_url_str)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_update_view(self):
        instance = StadiumFactory()
        instance.name = "Updated Instance"
        response = get_update_view(self.client, instance, StadiumModelSerializer, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_partial_update_view(self):
        instance = StadiumFactory()
        response = get_partial_view(self.client, instance.id, {"name": "Updated Instance"}, self.detail_url_str)
        instance.refresh_from_db()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["instance"]["name"], instance.name)

    def test_delete_view(self):
        instance = StadiumFactory()
        response = get_delete_view(self.client, self.detail_url_str, instance.pk)
        self.assertEqual(response.status_code, 204)
        self.assertFalse(Stadium.objects.filter(pk=instance.pk).exists())

    def test_ordering_fields(self):
        first_name, second_name, third_name = "Stadium A", "Stadium B", "Stadium C"
        StadiumFactory(name=second_name)
        StadiumFactory(name=first_name)
        StadiumFactory(name=third_name)

        response = self.client.get(self.list_url)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data["count"], Stadium.objects.count())
        self.assertEqual(response.data["results"][0]["name"], first_name)
        self.assertEqual(response.data["results"][1]["name"], second_name)
        self.assertEqual(response.data["results"][2]["name"], third_name)



from wbcore.frontend_user_configuration import FrontendUserConfiguration
from wbcore.metadata.configs.display.models import AppliedPreset, Preset
from wbcore.release_notes.models import ReleaseNote

from .base import DynamicModel, WBModel
from .fields import DynamicDecimalField, DynamicFloatField, YearField



from django.contrib import admin

from wbcore.contrib.workflow.models import (
    DecisionStep,
    EmailStep,
    FinishStep,
    JoinStep,
    ScriptStep,
    SplitStep,
    StartStep,
    UserStep,
)

STEP_BASE_FIELDS = (
    "id",
    "name",
    "workflow",
    "status",
    "code",
    "permission",
)


@admin.register(StartStep)
class StartStepAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = STEP_BASE_FIELDS


@admin.register(SplitStep)
class SplitStepAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = STEP_BASE_FIELDS


@admin.register(UserStep)
class UserStepAdmin(admin.ModelAdmin):
    search_fields = ("name", "assignee", "group")
    list_display = (
        "assignee",
        "group",
        "assignee_method",
        "notify_user",
        "display",
        "kwargs",
    ) + STEP_BASE_FIELDS


@admin.register(DecisionStep)
class DecisionStepAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = STEP_BASE_FIELDS


@admin.register(JoinStep)
class JoinStepAdmin(admin.ModelAdmin):
    search_fields = ("name", "wait_for_all")
    list_display = ("wait_for_all",) + STEP_BASE_FIELDS


@admin.register(ScriptStep)
class ScriptStepAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("script",) + STEP_BASE_FIELDS


@admin.register(EmailStep)
class EmailStepAdmin(admin.ModelAdmin):
    search_fields = (
        "to",
        "subject",
        "name",
    )
    list_display = ("subject", "template") + STEP_BASE_FIELDS
    fields = (
        "to",
        "subject",
        "name",
        "template",
        "cc",
        "bcc",
        "workflow",
        "status",
        "code",
        "permission",
    )


@admin.register(FinishStep)
class FinishStepAdmin(admin.ModelAdmin):
    search_fields = ("name",)
    list_display = ("write_preserved_instance",) + STEP_BASE_FIELDS



import pytest
from django.core import mail
from django.db.utils import IntegrityError
from faker import Faker
from pytest_mock import MockerFixture
from rest_framework.reverse import reverse
from rest_framework.test import APIRequestFactory

from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import User
from wbcore.contrib.authentication.viewsets.users import (
    activate_user,
    register_user,
    reset_password_email,
)
from wbcore.contrib.directory.models import Person

fake = Faker()


class TestUser:
    @pytest.fixture
    def test_user(self):
        return User()

    @pytest.fixture
    def test_user_data(self):
        return {
            "email": "john.neumann@intellineers.com",
            "password": "fake_password",
            "first_name": "John",
            "last_name": "Neumann",
        }

    def test_get_short_name(self, test_user_data):
        person = Person(first_name=test_user_data["first_name"], last_name=test_user_data["last_name"])
        user = User(profile=person)
        assert user.get_short_name() == test_user_data["first_name"]

    @pytest.mark.parametrize("is_valid", [True, False])
    def test_reset_password(self, is_valid: bool, mocker: MockerFixture, test_user, test_user_data):
        # Arrange
        mocker.patch.object(test_user, "email", test_user_data["email"])
        mock_password_reset_form = mocker.patch("django.contrib.auth.forms.PasswordResetForm")
        mock_instance = mock_password_reset_form.return_value
        mock_instance.is_valid.return_value = is_valid

        # Act
        test_user.reset_password(request="mocked_request")

        # Assert
        mock_password_reset_form.assert_called_once_with(data={"email": test_user_data["email"]})
        mock_instance.is_valid.assert_called_once()
        if is_valid:
            mock_instance.save.assert_called_once_with(
                request="mocked_request",
                email_template_name="password_reset_email.html",
                html_email_template_name="password_reset_email_html.html",
            )
        else:
            mock_instance.save.assert_not_called()

    def test_generate_and_verify_temporary_toke(self, mocker: MockerFixture, test_user):
        # Arrange
        other_user = User()
        # We need to mock an id, otherwise the generated tokens will be the same
        mocker.patch.object(test_user, "id", 1)
        mocker.patch.object(other_user, "id", 2)

        # Act
        user_token = test_user.generate_temporary_token()
        other_user_token = other_user.generate_temporary_token()

        # Assert
        assert test_user.verify_temporary_token(user_token)
        assert not test_user.verify_temporary_token(other_user_token)
        assert other_user.verify_temporary_token(other_user_token)
        assert not other_user.verify_temporary_token(user_token)

    @pytest.mark.django_db
    @pytest.mark.parametrize("first_name, last_name", [("John", "Neumann")])
    def test_generate_username(self, first_name: str, last_name: str, user_factory: UserFactory):
        username = User.generate_username(first_name, last_name)
        assert username == f"{first_name.lower()}-{last_name.lower()}"
        user_factory.create(username=username)
        username_1 = User.generate_username(first_name, last_name)
        assert username_1 == f"{first_name.lower()}-{last_name.lower()}-2"
        user_factory.create(username=username_1)
        assert User.generate_username(first_name, last_name) == f"{first_name.lower()}-{last_name.lower()}-3"

    def test_create_with_attributes(self, test_user_data, mocker: MockerFixture):
        # Arrange
        user_name = f"{test_user_data['first_name']}-{test_user_data['last_name']}"
        mock_user = mocker.Mock(spec=User)
        generate_user_name = mocker.patch(
            "wbcore.contrib.authentication.models.User.generate_username", return_value=user_name
        )
        create_user_mock = mocker.patch(
            "wbcore.contrib.authentication.models.User.objects.create_user", return_value=mock_user
        )

        # Act
        user = User.create_with_attributes(
            test_user_data["email"],
            test_user_data["password"],
            first_name=test_user_data["first_name"],
            last_name=test_user_data["last_name"],
        )

        # Assert
        assert user.check_password(test_user_data["password"])
        generate_user_name.assert_called_once_with(test_user_data["first_name"], test_user_data["last_name"])
        create_user_mock.assert_called_once_with(
            user_name,
            test_user_data["email"],
            test_user_data["password"],
            is_staff=False,
            is_superuser=False,
            is_active=False,
            is_register=False,
        )


class TestRegistrationAndActivationViews:
    @pytest.fixture
    def test_user(self):
        return User(
            password="fake_password",
            email="john.neumann@intellineers.com",
        )

    @pytest.fixture
    def registration_data(self, test_user):
        return {
            "email": test_user.email,
            "password": test_user.password,
            "first_name": "John",
            "last_name": "Neumann",
        }

    @pytest.fixture
    def missing_data_msg(self):
        return "email, first_name, last_name and password must be provided"

    @pytest.fixture
    def api_factory(self):
        return APIRequestFactory()

    @pytest.fixture
    def register_url(self):
        return reverse("wbcore:authentication:register")

    @pytest.mark.parametrize("should_fail", [True, False])
    def test_reset_password_with_user(self, should_fail, mocker: MockerFixture, test_user, api_factory):
        # Arrange
        mock_get_user = mocker.patch(
            "wbcore.contrib.authentication.models.users.UserManager.get",
            return_value=None if should_fail else test_user,
        )
        mock_reset_password = mocker.patch.object(test_user, "reset_password")
        url = reverse("wbcore:authentication:reset_password_email")

        # Act
        request = api_factory.post(url, {"email": test_user.email})
        response = reset_password_email(request)

        # Assert
        mock_get_user.assert_called_once_with(email="john.neumann@intellineers.com")

        if should_fail:
            mock_reset_password.assert_not_called()
        else:
            mock_reset_password.assert_called_once()
            assert response.data["status"] == "ok"
            assert (
                response.data["msg"]
                == "If the email matches a user, it will receive an email inviting him to reset his password."
            )
        assert response.status_code == 200

    @pytest.mark.parametrize("user_is_already_registered,", [True, False])
    def test_registration(
        self,
        user_is_already_registered,
        mocker: MockerFixture,
        test_user,
        api_factory,
        register_url,
        registration_data,
    ):
        # Arrange
        mocker.patch.object(test_user, "id", 1)
        patch_path = "wbcore.contrib.authentication.models.users.User.create_with_attributes"
        if user_is_already_registered:
            mock_create_with_attributes = mocker.patch(patch_path, side_effect=IntegrityError)
        else:
            mock_create_with_attributes = mocker.patch(patch_path, return_value=test_user)
        current_mailbox_count = len(mail.outbox)

        # Act
        request = api_factory.post(register_url, registration_data)
        response = register_user(request)

        # Assert
        mock_create_with_attributes.assert_called_once_with(
            registration_data["email"],
            registration_data["password"],
            first_name=registration_data["first_name"],
            last_name=registration_data["last_name"],
        )
        if user_is_already_registered:
            assert response.status_code == 409
            assert response.data["status"] == "fail"
            assert response.data["msg"] == "Your account already exists"
            assert len(mail.outbox) == current_mailbox_count
        else:
            assert response.status_code == 200
            assert response.data["status"] == "success"
            assert response.data["msg"] == "Please confirm your email address to complete the registration"
            assert len(mail.outbox) == current_mailbox_count + 1

            sent_mail = mail.outbox[-1]
            assert sent_mail.subject == "Activate your account."
            assert sent_mail.to == [test_user.email]

    @pytest.mark.parametrize("password", ["", "a" * 255])
    def test_registration_with_invalid_password(
        self, password, api_factory, register_url, registration_data, missing_data_msg
    ):
        # Arrange
        current_mailbox_count = len(mail.outbox)

        # Act
        request = api_factory.post(
            register_url,
            {
                "email": registration_data["email"],
                "password": password,
                "first_name": registration_data["first_name"],
                "last_name": registration_data["last_name"],
            },
        )
        response = register_user(request)

        # Assert
        assert response.status_code == 400
        assert response.data["status"] == "fail"
        if not password:
            assert response.data["msg"] == missing_data_msg
        else:
            assert (
                response.data["msg"] == "password is too long. please provider a password shorter than 128 characters."
            )
        assert len(mail.outbox) == current_mailbox_count

    @pytest.mark.parametrize("email", ["", "a" * 255 + "@test.com"])
    def test_registration_with_invalid_email(
        self, email, api_factory, register_url, registration_data, missing_data_msg
    ):
        # Arrange
        current_mailbox_count = len(mail.outbox)

        # Act
        request = api_factory.post(
            register_url,
            {
                "email": email,
                "password": registration_data["password"],
                "first_name": registration_data["first_name"],
                "last_name": registration_data["last_name"],
            },
        )
        response = register_user(request)

        # Assert
        assert response.status_code == 400
        assert response.data["status"] == "fail"
        if not email:
            assert response.data["msg"] == missing_data_msg
        else:
            assert response.data["msg"] == "email is too long. please provider an email shorter than 255 characters."
        assert len(mail.outbox) == current_mailbox_count

    @pytest.mark.parametrize(
        "first_name, last_name", [("", ""), ("a" * 101, ""), ("", "b" * 101), ("a" * 101, "b" * 101)]
    )
    def test_registration_with_invalid_name(
        self, first_name, last_name, api_factory, register_url, registration_data, missing_data_msg
    ):
        # Arrange
        current_mailbox_count = len(mail.outbox)

        # Act
        request = api_factory.post(
            register_url,
            {
                "email": registration_data["email"],
                "password": registration_data["password"],
                "first_name": first_name,
                "last_name": last_name,
            },
        )
        response = register_user(request)

        # Assert
        assert response.status_code == 400
        assert response.data["status"] == "fail"
        if len(first_name) > 100 or len(last_name) > 100:
            assert response.data["msg"] == "first and last name are too long. please provider a shorter name."
        else:
            assert response.data["msg"] == missing_data_msg
        assert len(mail.outbox) == current_mailbox_count

    @pytest.mark.parametrize("should_fail", [True, False])
    def test_activation(self, should_fail, mocker: MockerFixture, test_user, api_factory):
        # Assert
        test_uuid = "test_uuid"
        test_token = "wrong_token" if should_fail else test_user.generate_temporary_token()
        mocker.patch.object(test_user, "is_active", False)
        mocker.patch.object(test_user, "is_register", False)
        mocker.patch("wbcore.contrib.authentication.viewsets.users.get_object_or_404", return_value=test_user)
        test_user.save = mocker.MagicMock()
        url = reverse("wbcore:authentication:activate", args=[test_uuid, test_token])

        # Act
        request = api_factory.get(url)
        response = activate_user(request, test_uuid, test_token)

        # Assert
        if should_fail:
            assert response.status_code == 400
            assert test_user.is_active is False
            assert test_user.is_register is False
        else:
            assert response.status_code == 200
            assert test_user.is_active is True
            assert test_user.is_register is True
            test_user.save.assert_called_once()



from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
)
from wbcore.metadata.configs.preview import PreviewViewConfig


class GeographyPreviewConfig(PreviewViewConfig):
    def get_display(self) -> Display:
        return create_simple_display(
            [
                [
                    "name",
                    "parent",
                    "code_2",
                    "code_3",
                ]
            ]
        )



class Staticfiles:
    STATIC_URL = "static/"


class LocalStaticfiles(Staticfiles):
    STORAGES = {
        "default": {
            "BACKEND": "django.core.files.storage.FileSystemStorage",
        },
        "staticfiles": {
            "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage",
        },
    }

    @property
    def STATIC_ROOT(self):  # noqa
        return self.BASE_DIR.joinpath("staticfiles")


class S3Staticfiles(Staticfiles):
    STORAGES = {
        "default": {
            "BACKEND": "storages.backends.s3boto3.S3Boto3Storage",
        },
        "staticfiles": {
            "BACKEND": "storages.backends.s3boto3.S3Boto3Storage",
        },
    }



# -*- encoding: utf-8 -*-
import unicodedata
import uuid

from django.contrib.auth.models import AbstractBaseUser, BaseUserManager, PermissionsMixin
from django.contrib.auth.models import Group as DjangoBaseGroup
from django.contrib.auth.models import Permission as DjangoBasePermission
from django.contrib.auth.tokens import PasswordResetTokenGenerator
from django.db import models
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _


class UserManager(BaseUserManager):
    use_in_migrations = True

    @classmethod
    def normalize_username(cls, username):
        return unicodedata.normalize("NFKC", force_str(username))

    def _create_user(self, username, email, password, **extra_fields):
        if not email:
            raise ValueError(gettext("No E-Mail provided"))
        email = self.normalize_email(email)
        if not username:
            raise ValueError(gettext("No Username provided"))
        username = UserManager.normalize_username(username)
        user = self.model(email=email, username=username, **extra_fields)
        user.set_password(password)
        user.save(using=self._db)
        return user

    def create_user(self, username, email, password=None, **extra_fields):
        extra_fields.setdefault("is_staff", False)
        extra_fields.setdefault("is_superuser", False)
        extra_fields.setdefault("is_active", True)
        extra_fields.setdefault("is_register", True)
        return self._create_user(username, email.lower(), password, **extra_fields)

    def create_superuser(self, username, email, password, **extra_fields):
        extra_fields.setdefault("is_staff", True)
        extra_fields.setdefault("is_superuser", True)
        extra_fields.setdefault("is_active", True)
        extra_fields.setdefault("is_register", True)
        if extra_fields.get("is_staff") is not True:
            raise ValueError("Superuser must have is_staff=True.")
        if extra_fields.get("is_superuser") is not True:
            raise ValueError("Superuser must have is_superuser=True.")

        return self._create_user(username, email, password, **extra_fields)


class User(AbstractBaseUser, PermissionsMixin):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)
    email = models.EmailField(_("E-Mail"), unique=True)
    username = models.CharField(_("username"), max_length=255, unique=True)
    profile = models.OneToOneField("directory.Person", on_delete=models.PROTECT, related_name="user_account")
    is_staff = models.BooleanField(
        _("staff status"),
        default=False,
        help_text=_("Specifies whether the user can log into this admin site."),
    )
    is_active = models.BooleanField(
        _("active"),
        default=False,
        help_text=_(
            "Specifies whether this user should be treated as active. Unselect this instead of deleting accounts."
        ),
    )
    is_register = models.BooleanField(
        _("register"), default=True, help_text=_("Specifies whether this user has registered its email. ")
    )
    date_joined = models.DateTimeField(_("date joined"), default=timezone.now)
    metadata = models.JSONField(default=dict, blank=True)

    objects = UserManager()

    USERNAME_FIELD = "email"
    REQUIRED_FIELDS = ["username"]

    class Meta:
        verbose_name = _("user")
        verbose_name_plural = _("users")
        swappable = "AUTH_USER_MODEL"
        permissions = [("administrate_user", "Administrate Users"), ("is_internal_user", "Internal User")]

    @property
    def is_internal(self):
        from wbcore.permissions.registry import user_registry

        return self in user_registry.internal_users

    def get_full_name(self):
        full_name = "%s %s" % (self.profile.first_name, self.profile.last_name)
        return full_name.strip()

    def get_short_name(self):
        return self.profile.first_name

    def save(self, *args, **kwargs):
        from wbcore.contrib.directory.models.entries import Person

        self.username = self.username.lower()  # make sure username is lower case
        self.profile = Person.get_or_create_with_user(self)
        super().save(*args, **kwargs)

    def reset_password(self, request=None):
        """
        Create the Reset Password email including the new password to be used for the next login
        """
        from django.contrib.auth.forms import PasswordResetForm

        reset_password_form = PasswordResetForm(data={"email": self.email})
        if reset_password_form.is_valid():
            reset_password_form.save(
                request=request,
                email_template_name="password_reset_email.html",
                html_email_template_name="password_reset_email_html.html",
            )

    def generate_temporary_token(self):
        return PasswordResetTokenGenerator().make_token(self)

    def verify_temporary_token(self, token):
        return PasswordResetTokenGenerator().check_token(self, token)

    @classmethod
    def generate_username(cls, *names) -> str:
        username = "-".join(map(lambda x: x.lower(), names))
        similar = User.objects.filter(username__iregex=f"^({username}(-[0-9]*)*)$").count()
        return username if similar == 0 else f"{username}-{similar + 1}"

    @classmethod
    def create_with_attributes(cls, email, password, username=None, first_name=None, last_name=None):
        if not username and first_name and last_name:
            username = cls.generate_username(first_name, last_name)
        if not username:
            raise ValueError(gettext("You need to specify a username or a pair of first and last name"))
        return cls.objects.create_user(
            username, email, password, is_staff=False, is_superuser=False, is_active=False, is_register=False
        )

    @property
    def generic_auth_token_key(self) -> str | None:
        """
        Return generic valid useable token
        """
        token = self.auth_tokens.get_or_create(
            valid_until__isnull=True,
            protected_view_name__isnull=True,
            number_usage_left__isnull=True,
            defaults={
                "is_valid": True,
            },
        )[0]
        return token.key

    @property
    def first_name(self):
        return self.profile.first_name

    @property
    def last_name(self):
        return self.profile.last_name

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:authentication:user"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:authentication:userrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}}"


class Group(DjangoBaseGroup):
    class Meta:
        proxy = True

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:authentication:group"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:authentication:grouprepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"


class Permission(DjangoBasePermission):
    class Meta:
        proxy = True

    @classmethod
    def get_endpoint_basename(cls):
        return "wbcore:authentication:permission"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbcore:authentication:permissionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{name}}"



import pytest
from django.db import models
from pytest_mock import MockerFixture
from wbcore.contrib.workflow.models import (
    JoinStep,
    ProcessStep,
    Step,
    Transition,
)


class TestJoinStep:
    @pytest.fixture
    def join_step(self):
        return JoinStep()

    @pytest.fixture
    def mocked_step(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=Step, set_canceled=mocker.MagicMock())

    @pytest.fixture
    def mocked_process_step(self, mocker: MockerFixture, mocked_step):
        return mocker.MagicMock(spec=ProcessStep, step=mocked_step)

    @pytest.fixture
    def mocked_transition(self, mocker: MockerFixture):
        return mocker.MagicMock(spec=Transition)

    @pytest.fixture
    def mocked_execute_single_next_step(self, mocker: MockerFixture, join_step):
        return mocker.patch.object(join_step, "execute_single_next_step")

    def test_cancel_if_leading_to_self_transition_found(
        self,
        mocker: MockerFixture,
        join_step,
        mocked_step,
        mocked_process_step,
        mocked_transition,
    ):
        # Arrange
        mocker.patch.object(join_step, "id", 1)
        mocked_transition.to_step = join_step
        mocked_step.outgoing_transitions.all.return_value = [mocked_transition]
        # Act
        join_step.cancel_if_leading_to_self(mocked_step, mocked_process_step)
        # Assert
        mocked_step.set_canceled.assert_called_once_with(mocked_process_step)

    def test_cancel_if_leading_to_self_no_transition_found(
        self,
        mocker: MockerFixture,
        join_step,
        mocked_step,
        mocked_process_step,
        mocked_transition,
    ):
        # Arrange
        mocked_step.outgoing_transitions.all.return_value = [mocked_transition]
        # Act
        join_step.cancel_if_leading_to_self(mocked_step, mocked_process_step)
        # Assert
        mocked_step.set_canceled.assert_not_called()

    @pytest.mark.parametrize("with_transition_qs", [True, False])
    def test_run_with_wait_for_all(
        self,
        mocker: MockerFixture,
        join_step,
        mocked_process_step,
        with_transition_qs,
        mocked_execute_single_next_step,
    ):
        # Arrange
        mocker.patch.object(join_step, "wait_for_all", True)
        mocked_process_step.save = mocker.MagicMock()
        mocked_object_manager = mocker.patch("wbcore.contrib.workflow.models.transition.Transition.objects")
        mocked_qs = mocker.MagicMock(spec=models.QuerySet)
        mocked_qs.exists.return_value = with_transition_qs
        mocked_object_manager.filter.return_value = mocked_qs
        # Act
        join_step.run(mocked_process_step)
        # Assert
        assert mocked_process_step.state == ProcessStep.StepState.WAITING
        mocked_process_step.save.assert_called_once()
        if with_transition_qs:
            mocked_execute_single_next_step.assert_not_called()
        else:
            mocked_execute_single_next_step.assert_called_once_with(mocked_process_step)

    def test_run_without_wait_for_all(
        self,
        mocker: MockerFixture,
        join_step,
        mocked_process_step,
        mocked_execute_single_next_step,
    ):
        # Arrange
        mocker.patch.object(join_step, "wait_for_all", False)

        mocked_cancel_if_leading = mocker.patch.object(join_step, "cancel_if_leading_to_self")
        mocked_process_step.save = mocker.MagicMock()
        mocked_object_manager = mocker.patch("wbcore.contrib.workflow.models.step.ProcessStep.objects")
        unfinished_step_a = mocker.MagicMock(spec=ProcessStep, pk=123, step=mocker.MagicMock(spec=Step))

        unfinished_step_b = mocker.MagicMock(spec=ProcessStep, pk=456, step=mocker.MagicMock(spec=Step))

        mocked_qs = mocker.MagicMock(spec=models.QuerySet)
        mocked_qs.exclude.return_value = [unfinished_step_a, unfinished_step_b]
        mocked_object_manager.filter.return_value = mocked_qs
        # Act
        join_step.run(mocked_process_step)
        # Assert
        assert mocked_process_step.state == ProcessStep.StepState.WAITING
        assert mocked_cancel_if_leading.call_count == len(mocked_qs.exclude.return_value)
        mocked_process_step.save.assert_called_once()
        mocked_execute_single_next_step.assert_called_once()



from django.contrib.auth import get_user_model
from django.db import models


class Preset(models.Model):
    title = models.CharField(max_length=64)
    user = models.ForeignKey(
        to=get_user_model(), related_name="presets", on_delete=models.CASCADE, null=True, blank=True
    )
    display_identifier = models.CharField(max_length=512)
    display = models.JSONField(null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.title} - {self.user} ({self.display_identifier})"


class AppliedPreset(models.Model):
    user = models.ForeignKey(to=get_user_model(), related_name="applied_presets", on_delete=models.CASCADE)
    display_identifier_path = models.CharField(max_length=1024)
    preset = models.ForeignKey(
        to=Preset, related_name="applied_presets", on_delete=models.SET_NULL, null=True, blank=True
    )
    display = models.JSONField(null=True, blank=True)

    def __str__(self) -> str:
        return f"{self.display_identifier_path} ({self.user})"



from blockdiag.parser import ParseException
from django.core.exceptions import ValidationError
from django.http import HttpResponse
from django.template import exceptions
from markdown_blockdiag.utils import draw_blockdiag
from rest_framework import status
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView

from .models import Asset
from .template import render_template_for_templatetag


class BlockDiag(APIView):
    permission_classes = [IsAuthenticated]

    def post(self, request: Request) -> Response:
        try:
            blockdiag = request.data["markdown"]
            svg = draw_blockdiag(blockdiag, output_fmt="svg", font_antialias=True)
        except (ParseException, KeyError):
            return Response(status=status.HTTP_400_BAD_REQUEST)
        else:
            return Response(svg)


class TemplateTagView(APIView):
    permission_classes = []

    def post(self, request: Request) -> Response:
        if templatetag := request.data.get("templatetag"):
            try:
                return Response(render_template_for_templatetag(templatetag, request=request))
            except exceptions.TemplateSyntaxError:
                return Response("malformatted templatetag", status=400)
            except Exception:
                return Response("not handled exception", status=400)
        return Response("templatetag missing", status=400)


class AssetCreateView(APIView):
    permission_classes = []

    def post(self, request: Request) -> Response:
        try:
            asset = Asset.objects.create(file=request.data["file"])
            return Response(reverse("wbcore:asset-retrieve", args=[asset.id], request=request))
        except (KeyError, AttributeError):
            return Response("file missing", status=400)


class AssetRetrieveView(APIView):
    permission_classes = []

    def get(self, request: Request, uuid: str) -> Response:
        try:
            asset = Asset.objects.get(id=uuid)
            response = HttpResponse(asset.file, content_type=asset.content_type)
            response["Content-Disposition"] = f"attachment; filename={asset.file_url_name}"
            return response
        except Asset.DoesNotExist:
            return Response(status=404)
        except ValidationError as e:
            return Response(*e, status=400)
